from tkinter import Toplevel, Label, Widget, Event


def add_line_breaks(text:str, max_line_len:int, split_char=' ', replace_char='\n'):
    '''splits text into multiple lines - only splits at split_char
    the only way a line can be longer than max_line_len
    is if a single word is longer than max_line_len
    
    :return: str - text with some split chars replaced with '\n'
    '''
    output_text = ''
    line_len = 0
    for i, char in enumerate(text):
        if char == split_char:
            next_split = text[i + 1:].find(split_char) # chars until next split
            if next_split == -1: # no more split chars
                next_split = len(text[i + 1:])
            if next_split + line_len + 1 > max_line_len: # split here
                output_text += replace_char
                line_len = 0
                continue
        output_text += char
        line_len += 1
    return output_text


class ToolTip(Toplevel):
    ''' Popup to display information when the cursor hovers on a widget
    
        Info can be displayed on multiple lines by setting chars_per_line
        
        Popup will be displayed directly above the parent widget by default
        but it will check where the edge of the root frame is and adjust its
        position so that it does not extend beyond the root frame
    '''
    def __init__(self, master, chars_per_line=50, fade_inc:float=0.07,
                 fade_ms:int=20, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param chars_per_line: int - maximum characters before line break
            :param fade_inc: float - amount to adjust fade on every frame
            :param fade_ms: int (milliseconds) - time to wait before next frame
        '''
        self.__chars_per_line = chars_per_line
        self.__fade_inc, self.__fade_ms = fade_inc, fade_ms
        super().__init__(master)
        #make window invisible, on the top, and strip all window decorations/features
        self.attributes('-alpha', 0, '-topmost', True)
        self.overrideredirect(1)
        #style and create label. you can override style with kwargs
        style = dict(bd=2, relief='raised', font='courier 10 bold', bg='#FFFF99', anchor='w')
        self.label = Label(self, **{**style, **kwargs})
        self.label.grid(row=0, column=0, sticky='w')
        #used to determine if an opposing fade is already in progress
        self.fout:bool = False
        
    def bind(self, target:Widget, text:str, **kwargs):
        #bind Enter(mouseOver) and Leave(mouseOut) events to the target of this tooltip
        target.bind('<Enter>', lambda e: self.fadein(0, text, e))
        target.bind('<Leave>', lambda e: self.fadeout(1 - self.__fade_inc, e))

    def fadein(self, alpha:float, text:str=None, event:Event=None,
               widget_pos:'tuple[int]'=None):
        ''':param widget_pos: tuple[int] - (width, height, x, y)'''
        #if event and text then this call came from target
        #~ we can consider this a "fresh/new" call
        if event is not None and text is not None:
            if self.fout: # if we are in the middle of fading out jump to end of fade
                self.attributes('-alpha', 0)
                self.fout = False # indicate that we are fading in
            if '\n' not in text: # if so, assume line breaks have already been done
                text = add_line_breaks(text, self.__chars_per_line, replace_char=' \n ') # to pad each line with space
            self.label.configure(text=f'{text:^{len(text)+2}}') # pad with a space on either side
            self.update() # update so the proceeding geometry will be correct

            # check widget position (widget that calls tool_tip popup such as button or label)
            if widget_pos is not None:
                widget_width, widget_height, widget_x, widget_y = widget_pos
            else:
                widget_width = event.widget.winfo_width()
                widget_height = event.widget.winfo_height()
                widget_x = event.widget.winfo_rootx()
                widget_y = event.widget.winfo_rooty()

            # compute popup geometry
            w = self.label.winfo_width()
            h = self.label.winfo_height()
            x = widget_x + int((widget_width - w) / 2) # to center horizontally
            y = widget_y - h # directly above widget

            # check root coordinates
            root_width = event.widget.winfo_toplevel().winfo_width()
            root_height = event.widget.winfo_toplevel().winfo_height()
            root_x = event.widget.winfo_toplevel().winfo_rootx() # x position of app on computer screen
            root_y = event.widget.winfo_toplevel().winfo_rooty() # y position of app on computer screen

            # adjust so that popup is entirely within the root window
            if x + w > root_x + root_width: # enforce right edge
                x = root_x + root_width - w
            if y + h > root_y + root_height: # enforce bottom edge
                y = root_y + root_height - h
            x = max(root_x, x) # enforce left edge
            y = max(root_y, y) # enforce top edge

            # ensure tool_tip is not on top of widget - this makes a button unclickable!
            # this can only happen if tool_tip was moved down by top enforcement, so move it to below widget
            if y + h > widget_y:
                y = widget_y + widget_height # position tool_tip below widget

            #apply geometry
            self.geometry(f'{w}x{h}+{x}+{y}')
               
        #if we aren't fading out, fade in
        if not self.fout:
            self.attributes('-alpha', alpha)
        
            if alpha < 1:
                self.after(self.__fade_ms, lambda: self.fadein(min(alpha + self.__fade_inc, 1)))

    def fadeout(self, alpha:float, event:Event=None):
        #if event then this call came from target 
        #~ we can consider this a "fresh/new" call
        if event is not None:
            #indicate that we are fading out
            self.fout = True
        
        #if we aren't fading in, fade out        
        if self.fout:
            self.attributes('-alpha', alpha)
        
            if alpha > 0:
                self.after(self.__fade_ms, lambda: self.fadeout(max(alpha - self.__fade_inc, 0)))

    def set_text(self, text:str):
        '''sets popup text - to be called while popup is visible
        for example, when a button is clicked'''
        if '\n' not in text: # if so, assume line breaks have already been done
            text = add_line_breaks(text, self.__chars_per_line, replace_char=' \n ') # to pad each line with space
        self.label.configure(text=f'{text:^{len(text)+2}}') # pad with a space on either side
        #update so the proceeding geometry will be correct
        self.update()
        w = self.label.winfo_width()
        h = self.label.winfo_height()
        #apply geometry - the position (x, y) is already set - only need to update height/width
        self.geometry(f'{w}x{h}')



from tkinter import Frame, Label
from datetime import datetime

from .timer import Timer


class TempLabel(Label):
    ''' Simple Tkinter Label that clears after a certain length of time
    '''
    def __init__(self, master:Frame, duration=5.0, default_text='', **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param duration: float (seconds) - time to wait before clearing
            :param default_text: str - text displayed when clearing
        '''
        self.__default_text = default_text
        self.__text_history: list[dict] = []
        super().__init__(master, text=self.__default_text, **kwargs)

        self.__delay = 0.1 # arbitrary time between steps
        self.__Timer = Timer(self.__delay, callback=lambda s: None,
                             end_callback=self.__to_default,
                             max_step=int(duration / self.__delay))
        
    def __to_default(self):
        '''called when timer reaches the end - reset label'''
        self.config(text=self.__default_text)

    def set_duration(self, duration:float):
        '''updates label clear duration (seconds)'''
        self.__Timer.set_max_step(int(duration / self.__delay))

    def set_text(self, text:str, **kwargs):
        '''
        Purpose
        -------
            Updates label text and any other parameters such as foregroud color
            Resets the timer
        '''
        self.__text_history.append({'text':text, 'time':datetime.now()})
        self.config(text=text, **kwargs)
        self.__Timer.reset()
        self.__Timer.start() # in case label was on default (Timer not running)

    def clear_history(self):
        '''clears text history'''
        self.__text_history = []

    def get_history(self) -> list:
        '''
        Purpose
        -------
            collects all text updates, with their corresponding timestamps
        
        Returns
        -------
            :return: list[dict] - each dict contains keys:
                text: str - text that was passed to set_text()
                time: datetime object - timestamp when set_text() was called
        '''
        return self.__text_history




from tkinter import Menu, Frame, Label
from fractions import Fraction


class OptionMenu(Label):
    ''' Label that displays a tkinter Menu when clicked

        OptionMenu was created because tk.OptionMenu and ttk.OptionMenu are not
        aesthetically pleasing and they lack customizable features
    
        OptionMenu.set() must be called when an option is selected in order to
        update the label
    '''
    def __init__(self, master:Frame, menu:Menu, default='', bg:str='#ffffff',
                 fg='#000000', hover_bg:str='#aaaaaa',
                 font_name='Segoe UI', font_size=10, active=True):
        '''        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param menu: tk.Menu - displayed when label is clicked
            :param default: str - initial text displayed on label
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - text color
            :param hover_bg: str (hex code) - background color when cursor hovers
            :param active: bool - if False, cursor click will not do anything
        '''
        self.menu = menu
        self.text = default
        self.bg = bg
        self.hover_bg = hover_bg
        self.__active = active
        Label.__init__(self, master, text=self.text, font=(font_name, font_size), fg=fg, bg=bg)
        self.bind('<Enter>', self.__hover_enter)
        self.bind('<Leave>', self.__hover_leave)
        self.bind('<Button-1>', self.__popup)

    def set_active(self):
        '''sets state to active so that menu can be viewed and edited by user'''
        self.__active = True

    def set_inactive(self):
        '''sets state to inactive so that hover and click events don't do anything'''
        self.__active = False

    def __hover_enter(self, event=None):
        '''called when cursor enters label'''
        if self.__active:
            self.config(bg=self.hover_bg)

    def __hover_leave(self, event=None):
        '''called when cursor leaves label'''
        if self.__active:
            self.config(bg=self.bg)

    def __popup(self, event=None):
        '''called when cursor clicks on label'''
        if self.__active:
            self.menu.tk_popup(self.winfo_rootx(), self.winfo_rooty() + self.winfo_height())

    def set(self, text:str):
        '''called internally or when a dropdown option is clicked by user'''
        self.text = text
        self.config(text=self.text)

    def get_text(self) -> str:
        '''returns currently selected option'''
        return self.text

class BasicDropDown(OptionMenu):
    ''' Dropdown menu to allow user to selected from a list of options
    
        When an option is selected, calls callback function with the single
        str argument being the option that was selected
    '''
    def __init__(self, master, options:list, default:str, callback=None,
                 bg:str='#ffffff', fg:str='#000000', hover_bg:str='#aaaaaa',
                 active=True):
        '''dropdown menu containing list of str
        
        Parameters
        ----------
            :param master: tk.Frame - frame in which to put dropdown menu
            :param options: list of str - dropdown options available to user
            :param default: str - default selection - should be in options but not mandatory
            :param callback: function (str) - called when an option is selected
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - foreground color
            :param hover_bg: str (hex code) - background color when cursor hovers
            :param active: bool - if True, dropdown menu is editable by user
        '''
        self._callback = callback
        menu = Menu(master, tearoff=False, bg=bg, fg=fg, activebackground=hover_bg)
        for op in options:
            menu.add_command(label=op, command=lambda x=op: self.update(x, callback=True))

        OptionMenu.__init__(self, master, menu, default=default, bg=bg, fg=fg,
                            hover_bg=hover_bg, active=active)

    def update(self, x:str, callback=False):
        '''called when a dropdown item is selected by user or called internally to set value'''
        self.set(x)
        if callback and self._callback:
            self._callback(self.get())

    def get(self):
        '''returns currently selected option'''
        return self.get_text()

class NumberDropDown(BasicDropDown):
    ''' Special version of BasicDropDown to deal with integers and floats
    
        .get() method returns an integer or float depending on 'return_type'
    '''
    def __init__(self, master, options:list, default:int, return_type='int',
                 **kwargs):
        '''dropdown for selecting number from options
        
        Parameters
        ----------
            :param master : tk.Frame - frame in which to place drowdown menu
            :param options : list[int | float] - dropdown options
            :param default : int - default selection, should be in options but not mandatory
            :param return_type : Literal['int', 'float'] - type to be returned by self.get
            **kwargs includes: ['bg', 'fg', 'hover_bg', 'active']
        '''
        BasicDropDown.__init__(self, master, [str(x) for x in options],
                               str(default), **kwargs)
        self.__return_type = return_type

    def update(self, x:str, callback=False):
        '''called when a dropdown item is selected by user or called internally'''
        self.set(str(x))
        if callback and self._callback:
            self._callback(self.get())
    
    def get(self):
        '''returns currently selected option'''
        if self.__return_type == 'float':
            return float(self.get_text())
        else:
            return int(self.get_text())

class TuningDropDown(OptionMenu):
    ''' Dropdown menu to select tuning or create a custom tuning
    
        Tuning can be retrieved with .get() and .get_labels() either as
        a list of note indices such as [19, 24, 29, 34, 38, 43]
        or a list of labels such as ['E', 'A', 'D', 'G', 'B', 'E']
    '''
    def __init__(self, master, default_tuning:list, callback=None,
                 bg:str='#ffffff', fg='#000000', hover_bg:str='#aaaaaa',
                 active=True):
        '''dropdown to select tuning - default must be in tuning keys'''
        self._callback = callback
        self.tunings = {'E Standard':[19, 24, 29, 34, 38, 43],
                        'E♭ Standard':[18, 23, 28, 33, 37, 42],
                        'D Standard':[17, 22, 27, 32, 36, 41],
                        'Drop D':[17, 24, 29, 34, 38, 43],
                        'Drop D♭':[16, 23, 28, 33, 37, 42],
                        'Drop C':[15, 22, 27, 32, 36, 41]}
        self.labels = {'E Standard':['E', 'A', 'D', 'G', 'B', 'E'],
                        'E♭ Standard':['D♯', 'G♯', 'C♯', 'F♯', 'A♯', 'D♯'],
                        'D Standard':['D', 'G', 'C', 'F', 'A', 'D'],
                        'Drop D':['D', 'A', 'D', 'G', 'B', 'E'],
                        'Drop D♭':['C♯', 'G♯', 'C♯', 'F♯', 'A♯', 'D♯'],
                        'Drop C':['C', 'G', 'C', 'F', 'A', 'D']}
        menu = Menu(master, tearoff=False, bg=bg, fg=fg, activebackground=hover_bg)
        for tuning in self.tunings.keys():
            menu.add_command(label=tuning, command=lambda x=tuning: self.__update(x))
        menu.add_command(label='Custom Tuning', command=self.set_custom_tuning)

        OptionMenu.__init__(self, master, menu, bg=bg, fg=fg,
                            hover_bg=hover_bg, active=active)
        self.set_tuning(default_tuning)

    def set_custom_tuning(self):
        '''open window to create custom tuning'''

    def __update(self, x:str):
        '''called when dropdown is selected by user'''
        self.set(x)
        if self._callback:
            self._callback(self.get())

    def set_tuning(self, tuning:list):
        '''sets tuning based on list of note indices
        intended to be called only when a new project is loaded
        
        Parameters
        ----------
            :param tuning: list[int] - note indices of each string, lowest first
        '''
        new_tuning = 'E Standard' # in case no match is found
        for text, inds in self.tunings.items():
            if tuning == inds:
                new_tuning = text
                break
        self.set(new_tuning)

    def get(self):
        '''returns tuning as list of note indices (ints) such as: [19, 23, 29, 34, 38, 43]'''
        return self.tunings[self.get_text()]

    def get_labels(self):
        '''returns tuning labels such as: ['E', 'A', 'D', 'G', 'B', 'E']'''
        return self.labels[self.get_text()]

class KeyDropDown(OptionMenu):
    ''' Dropdown menu to edit the key of a song
    
        Key is displayed to user in the common form 'A' or 'C#'
        but the .get() method returns the key in lilypond format such as
        'a' or 'cis' or 'bes'
    '''
    def __init__(self, master, default:str, callback=None, bg:str='#ffffff',
                 fg:str='#000000', hover_bg:str='#aaaaaa', active:bool=True):
        '''
        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param default: str - default key in lilypond format
            :param bg: str - background color
            :param fg: str - foreground color
            :param hover_bg: str - background color when cursor is hovering
            :param active: bool - if True, dropdown is editable by user
        '''
        self._callback = callback
        sharp_sign_unicode, flat_sign_unicode, natural_sign_unicode = '\u266F', '\u266D', '\u266E'
        self.keys_dict = {'A':'a',
                          f'A{sharp_sign_unicode}':'ais',
                          f'A{flat_sign_unicode}':'aes',
                          'B':'b',
                          f'B{flat_sign_unicode}':'bes',
                          'C':'c',
                          f'C{sharp_sign_unicode}':'cis',
                          'D':'d',
                          f'D{sharp_sign_unicode}':'dis',
                          f'D{flat_sign_unicode}':'des',
                          'E':'e',
                          f'E{flat_sign_unicode}':'ees',
                          'F':'f',
                          f'F{sharp_sign_unicode}':'fis',
                          'G':'g',
                          f'G{flat_sign_unicode}':'ges',
                          f'G{sharp_sign_unicode}':'gis'}

        keys = list(self.keys_dict.keys())
        default_key = 'A'
        for key in keys:
            if self.keys_dict[key] == default:
                default_key = key
                break
        
        menu = Menu(master, tearoff=False, bg=bg, fg=fg, activebackground=hover_bg)
        for root_key in ['A', 'B', 'C', 'D', 'E', 'F', 'G']:
            inds = [i for i in range(len(keys)) if root_key in keys[i]]
            sub_menu = Menu(menu, tearoff=False, bg=bg, fg=fg, activebackground=hover_bg)
            menu.add_cascade(label=root_key, menu=sub_menu)
            for i in inds:
                sub_menu.add_command(label=keys[i], command=lambda x=keys[i]: self.__update(x))

        OptionMenu.__init__(self, master, menu, default=default_key, bg=bg,
                            fg=fg, hover_bg=hover_bg, active=active)

    def __update(self, x:str):
        '''called when key is changed by user by clicking dropdown menu'''
        self.set(x)
        if self._callback:
            self._callback(self.get())

    def set_key(self, new_key:str, callback=False):
        '''set the key - called externally to KeyDropDown class
        
        Parameters
        ----------
            :param new_key:str - key in lilypond format such as 'a' or 'cis' or 'bes'
            :param callback: bool - if True, calls callback function
        '''
        key_name = 'A'
        for key in list(self.keys_dict.keys()):
            if self.keys_dict[key] == new_key:
                key_name = key
                break
        self.set(key_name)
        if callback and self._callback:
            self._callback(self.get())

    def get(self) -> str:
        '''returns key in lilypond format such as "ais" or "des" or "b"'''
        return self.keys_dict[self.get_text()]

class MeterDropDown(Frame):
    ''' Dropdown menu to edit the time signature of a song
    
        Numerator of time signature must be > 0 and denominator must be a
        power of 2 such as 4, 8, or 16
    '''
    def __init__(self, master, callback=None, default=(4, 4),
                 bg:str='#ffffff', fg:str='#000000', **kwargs):
        '''dropdown for selecting meter
        
        Parameters
        ----------
            :param master: tk.Frame - frame in which to place drowdown menu
            :param callback: function (str) - called when meter is changed
            :param default: tuple[int, int] - default meter such as (4, 4)
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - foreground color
        '''
        self._callback = callback
        Frame.__init__(self, master, bg=bg)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=0)
        self.grid_columnconfigure(2, weight=1)
        Label(self, text='/', bg=bg, fg=fg).grid(row=0, column=1, padx=2)
        self.__numerator = NumberDropDown(self, list(range(1, 17)), default[0],
                                          callback=self.__update, bg=bg,
                                          fg=fg, **kwargs)
        self.__numerator.grid(row=0, column=0, sticky='nsew')
        self.__denominator = NumberDropDown(self, [1, 2, 4, 8, 16, 32],
                                            default[1], callback=self.__update,
                                            bg=bg, fg=fg, **kwargs)
        self.__denominator.grid(row=0, column=2, sticky='nsew')

    def set_active(self):
        self.__numerator.set_active()
        self.__denominator.set_active()

    def set_inactive(self):
        self.__numerator.set_inactive()
        self.__denominator.set_inactive()

    def __update(self, x=None):
        '''called when the numerator or denominator is changed'''
        if self._callback:
            self._callback(self.get_str())

    def set_meter(self, numerator:int, denominator:int, callback:bool=False):
        '''called externally to change meter
        
        Parameters
        ----------
            :param numerator: int - top of time signature - must be >= 0
            :param denominator: int - bottom of time signature - must be power of 2
            :param callback: int - if True, calls callback function
        '''
        assert isinstance(numerator, int), f'Meter numerator must be an int, not {type(numerator)}'
        assert isinstance(denominator, int), f'Meter denominator must be an integer, not {type(denominator)}'
        assert numerator > 0, f'Meter numerator must be greater than 0, not {numerator}'
        assert denominator in [1, 2, 4, 8, 16, 32], f'Meter denominator must be a power of 2, not {denominator}'
        self.__numerator.update(numerator, callback=False)
        self.__denominator.update(denominator, callback=False)
        if callback:
            self.__update()

    def get_str(self) -> str:
        '''returns meter in form 4/4 or 6/8'''
        return f'{self.__numerator.get()}/{self.__denominator.get()}'
    
    def get_fraction(self) -> Fraction:
        '''returns meter as an un-reduced Fraction'''
        return Fraction(self.__numerator.get(), self.__denominator.get(),
                        _normalize=False)



import tkinter as tk
import tkinter.ttk as ttk


class ScrollableFrame(tk.Frame):
    ''' Frame that allows widgets to be placed outside the visible area and
        scrolled into view
        
        <MouseWheel> event must be bound to all widgets placed in the
        scrollable frame. By default, the <MouseWheel> event is only bound
        to the frame itself
        
        All widgets must be put in self.scrollable_frame, not self
    '''
    def __init__(self, master:tk.Frame, bg:str, include_scrollbar:bool=True,
                 scrollbar_side='right', yscrollincrement:int=1,
                 check_hover=False, *args, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param bg: str (hex code) - background color
            :param include_scrollbar: bool - if True, scrollbar is packed right
            :param scrollbar_side: str - Literal['left', 'right']
            :param yscrollincrement: int - pixels scrolled per mouse scroll (70x)
            :param check_hover: bool - if True, will only respond to mousewheel
                                       events if the cursor is hovering on Frame
        '''
        assert scrollbar_side in ['left', 'right'], f'Invalide scrollbar side: {scrollbar_side}'
        self.hovering = not check_hover
        super().__init__(master, *args, **kwargs)
        style = ttk.Style(master)
        style.theme_use("alt")
        style.layout('arrowless.Vertical.TScrollbar',
                     [('Vertical.Scrollbar.trough', 
                     {'children': [('Vertical.Scrollbar.thumb',
                                    {'expand': '1', 'sticky': 'nswe'})],
                                    'sticky': 'ns'})])
        self.canvas = tk.Canvas(self, bg=bg, highlightthickness=0,
                                yscrollincrement=yscrollincrement)
        self.scrollable_frame = tk.Frame(self.canvas, bg=bg)
        self.scrollable_frame.bind("<Configure>", self.OnFrameConfigure)
        self.canvas_id = self.canvas.create_window((0, 0), window=self.scrollable_frame, anchor="nw")
        if include_scrollbar:
            scrollbar = ttk.Scrollbar(self, orient="vertical", command=self.canvas.yview,
                                      style='arrowless.Vertical.TScrollbar')
            self.canvas.configure(yscrollcommand=scrollbar.set)
            scrollbar.pack(side=scrollbar_side, fill="y")
        self.canvas.pack(side="left", fill="both", expand=True)
        style.configure("arrowless.Vertical.TScrollbar", troughcolor=bg)
        self.canvas.bind('<Configure>', self.FrameWidth)
        self.canvas.bind("<MouseWheel>", self.on_mousewheel)
        self.scrollable_frame.bind("<MouseWheel>", self.on_mousewheel)
        if check_hover:
            self.bind('<Enter>', self.hover_enter)
            self.bind('<Leave>', self.hover_leave)

    def on_mousewheel(self, event, scroll_px=70):
        '''called when mouse wheel moves
        does not allow canvas to scroll past the top'''
        if self.hovering:
            scrollby = int(max(-self.canvas.canvasy(0), -event.delta * scroll_px /120))
            self.canvas.yview_scroll(scrollby, "units")

    def scroll_to_top(self):
        '''scrolls to the top of page'''
        self.canvas.yview_scroll(int(-self.canvas.canvasy(0)), "units")

    def hover_enter(self, event=None):
        '''called when the cursor enters frame'''
        self.hovering = True

    def hover_leave(self, event=None):
        '''called when the cursor leaves frame'''
        self.hovering = False

    def FrameWidth(self, event):
        '''called when canvas is resized'''
        self.canvas.itemconfig(self.canvas_id, width = event.width)

    def OnFrameConfigure(self, event):
        '''called when scrollable frame is resized'''
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))



from tkinter import Canvas, Frame, Label


class ProgressBar(Frame):
    ''' Progress bar that displays completion percentage and a label
    
        Completion percentage is computed in terms of iterations so that
        the progress bar can be easily increment with each iteration of a
        function using the .increment() method
    '''
    def __init__(self, master:Frame, iterations:int, bg:str, inactive_color=None,
                    active_color:str='#00ff00', text_color:str='#000000',
                    border_color:str='#ffffff', text='Loading...',
                    border_thickness=1, width=400, height=40, start_value:int=0,
                    text_font_name='Segoe UI bold', val_font_name='Segoe UI',
                    text_font_size=14, val_font_size=12, decimals:int=0,
                    text_side='bottom', center=True):
        '''Simple progress bar that displays completion percentage
        
        Parameters
        ----------
            :param master: tk.Frame - widget in which to pack progress bar
            :param iterations: int - number of iterations until completion
            :param bg: str (hex code) -  frame background color
            :param inactive_color: str (hex code) - color of uncompleted part of progress bar
            :param active_color: str (hex code) - color of completed part of progress bar
            :param text_color: str (hex code) - color of percentage label
            :param border_color: str (hex code) - color of canvas border
            :param text: str - text displayed while progress bar is running
            :param border_thickness: int - thickness of progress bar border
            :param width: int (pixels) - width of progress bar
            :param height: int (pixels) - height of progress bar
            :param start_value: int - iterations already complete
            :param decimals: int - number of decimal places to display on percentage
            :param text_side: str - where to pack loading text relative to progress bar
            :param center: bool - if True, progress bar will be placed within self
        '''
        Frame.__init__(self, master, bg=bg)
        self.width, self.height = width, height
        self.iterations = max(1, iterations)
        self.current_value = start_value
        self.decimals = decimals
        self.bd = border_thickness
        inactive_color = inactive_color if inactive_color else bg

        self.frame = Frame(self, bg=bg)
        if center:
            self.frame.place(relx=0.5, rely=0.5, anchor='center')
        else:
            self.frame.pack(side='top')
        self.loading_label = Label(self.frame, text=text, bg=bg, fg=text_color,
                                   font=(text_font_name, text_font_size))
        self.loading_label.pack(side=text_side)
        self.canvas = Canvas(self.frame, bg=inactive_color, width=self.width,
                             height=self.height, highlightthickness=self.bd,
                             highlightbackground=border_color)
        self.canvas.pack(side='left')
        self.label = Label(self.frame, text=self.get_percentage_text(), bg=bg,
                           fg=text_color, font=(val_font_name, val_font_size))
        self.label.pack(side='left')
        self.bar_id = self.canvas.create_rectangle(*self.get_bar_coords(),
                                                   fill=active_color, width=0,
                                                   state='normal')

    def set(self, perc:float):
        '''sets progress bar given a completion percentage'''
        assert perc >= 0 and perc <= 1, f'Progress Bar Error: Invalid completion percentage: {perc}'
        self.current_value = int(self.iterations * perc)
        self.canvas.coords(self.bar_id, *self.get_bar_coords())
        self.label.config(text=self.get_percentage_text())

    def set_text(self, text:str):
        '''updates progress bar loading text'''
        self.loading_label.config(text=text)

    def get_percentage_text(self) -> str:
        '''return completion percentage'''
        perc = self.current_value * 100 / self.iterations
        if self.decimals == 0:
            return str(int(perc)) + '%'
        text = str(round(perc, self.decimals))
        if '.' not in text:
            return text + '.' + '0' * self.decimals
        return  text + '0' * (self.decimals - len(text.split('.')[1])) + '%'

    def get_bar_coords(self):
        '''returns x0, y0, x1, y1 for active bar based on self.current_value'''
        x0, x1 = self.bd, self.width * self.current_value / self.iterations + self.bd
        y0, y1 = self.bd, self.height + self.bd
        return x0, y0, x1, y1

    def reset(self, iterations=None, current_value=None):
        '''reset iterations and/or current value'''
        if iterations != None:
            self.iterations = iterations
        if current_value != None:
            self.current_value = current_value
        else:
            self.current_value = 0
        self.canvas.coords(self.bar_id, *self.get_bar_coords())
        self.label.config(text=self.get_percentage_text())

    def increment(self, inc=1):
        '''called everytime an iteration is completed'''
        self.current_value = max(0, min(self.current_value + inc, self.iterations))
        self.canvas.coords(self.bar_id, *self.get_bar_coords())
        self.label.config(text=self.get_percentage_text())





from tkinter import Frame, Label

from PIL import Image, ImageTk
import numpy as np
import cv2, os

from .icons import icons
from .buttons import image_replace_colors

__all__ = ['Icon', 'CheckIcon', 'CheckLabel', 'IconCheckLabel']


class Icon(Label):
    ''' Displays icon from png file with custom background and foreground color
    
        png file MUST have black background with white icon
    '''
    def __init__(self, master:Frame, icon_path:str, bg:str='#000000', fg:str='#ffffff',
                 bg2=None, fg2=None, h:int=24, w:int=24, status=False):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param icon_path: str or np.array - path to .png file or image array
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - foreground color
            :param bg2: str (hex code) - optional secondary background color
            :param fg2: str (hex code) - optional secondary foreground color
            :param h: int (pixels) - icon height
            :param w: int (pixels) - icon width
            :param status: bool - False is bg and True is bg2
        '''        
        super().__init__(master, bg=bg)

        bg2 = bg2 if bg2 is not None else bg
        fg2 = fg2 if fg2 is not None else fg

        img_ar = self.__get_img_array(icon_path)
        self.img0 = cv2.resize(image_replace_colors(img_ar, [('#ffffff', fg), ('#000000', bg)]), (w, h))
        self.img1 = cv2.resize(image_replace_colors(img_ar, [('#ffffff', fg2), ('#000000', bg2)]), (w, h))

        # TODO: can't get img1 colors to show up when calling set()

        self.set(status)

    def __get_img_array(self, icon_path):
        '''returns image path as numpy array
        :icon_path: str (filepath) or np.array
        '''
        if isinstance(icon_path, str): # path to image
            assert len(icon_path) > 4, f'Icon Error: Invalid path: {icon_path}'
            assert icon_path[-4:] == '.png', f'Icon Error: icon_path is not a .png file: {icon_path}'
            assert os.path.exists(icon_path), f'Icon Error: Path to .png file does not exist: {icon_path}'
            return cv2.imread(icon_path)
        elif isinstance(icon_path, np.ndarray): # 3d numpy array
            return icon_path
        else:
            raise TypeError(f'Invalid icon input to Icon: {icon_path}')
        
    def set(self, status:bool):
        '''sets background color'''
        if status: # bg2
            image = ImageTk.PhotoImage(image=Image.fromarray(self.img1), master=self)
        else:
            image = ImageTk.PhotoImage(image=Image.fromarray(self.img0), master=self)
        self.config(image=image)
        self.image = image

class ToggleIcon(Frame):
    ''' Displays one of two icons depending on status
    
        Status can be set and retrieved with .set() and .get() methods

        Status 0 (False) is the first icon and status 1 (True) is second icon

        Do not use directly - only used for inheritance by CheckIcon and CheckLabel
    '''
    def __init__(self, master:Frame, path1:str, path2:str,
                 inactive_bg:str='#000000', active_bg:str=None,
                 inactive_fg:str='#888888', active_fg:str=None,
                 h:int=24, w:int=24, selected=False):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param Icon0: Icon - first icon (status 0 or False)
            :param Icon1: Icon - second icon (status 1 or True)
            :param status: bool - initial status
            :param bg: str (hex code) - background color
        '''
        super().__init__(master, bg=inactive_bg)
        self.__status = selected

        active_bg = active_bg if active_bg is not None else inactive_bg
        active_fg = active_fg if active_fg is not None else inactive_fg
        self.__icon0 = Icon(self, path1, bg=inactive_bg, bg2=active_bg, fg=inactive_fg, h=h, w=w)
        self.__icon1 = Icon(self, path2, bg=inactive_bg, bg2=active_bg, fg=active_fg, h=h, w=w)

        self.set(self.__status)

    def bindall(self, *args, **kwargs):
        '''binds to all sub widgets'''
        self.bind(*args, **kwargs)
        self.__icon0.bind(*args, **kwargs)
        self.__icon1.bind(*args, **kwargs)

    def set(self, status:bool):
        '''sets checked status - True for checked, False for unchecked'''
        self.__status = status
        if status: # set to checked (icon1)
            self.__icon0.pack_forget()
            self.__icon1.pack(fill='both')
        else: # set to unchecked (icon0)
            self.__icon1.pack_forget()
            self.__icon0.pack(fill='both')

    def set_bg(self, status:bool):
        '''sets main (False) or alternate (True) background'''
        self.__icon0.set(status)
        self.__icon1.set(status)

    def get(self) -> bool:
        '''returns current checked status'''
        return self.__status

class CheckIcon(ToggleIcon):
    ''' Displays checked or unchecked icon
    
        Checked status can be set and retrieved with .set() and .get() methods
    '''
    def __init__(self, master:Frame, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param inactive_bg: str (hex code) - background color when unchecked
            :param active_bg: str (hex code) or None - check background color (if different from inactive_bg)
            :param inactive_fg: str (hex code) - foreground color when unchecked
            :param active_fg: str (hex code) - foreground color when checked (if different from inactive_fg)
            :param h: int (pixels) - icon height
            :param w: int (pixels) - icon width
            :param selected: bool - initial selection status
        '''
        super().__init__(master, icons['box'], icons['checkbox'], **kwargs)

class CheckLabel(ToggleIcon):
    ''' Displays checkmark or blank for unchecked

        Sames as CheckIcon but without the box
    
        Checked status can be set and retrieved with .set() and .get() methods
    '''
    def __init__(self, master:Frame, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param inactive_bg: str (hex code) - background color when unchecked
            :param active_bg: str (hex code) or None - check background color (if different from inactive_bg)
            :param inactive_fg: str (hex code) - foreground color when unchecked
            :param active_fg: str (hex code) - foreground color when checked (if different from inactive_fg)
            :param h: int (pixels) - icon height
            :param w: int (pixels) - icon width
            :param selected: bool - initial selection status
        '''
        super().__init__(master, np.zeros((24, 24, 3)).astype(np.uint8), icons['check'], **kwargs)

class IconCheckLabel(Frame):
    ''' Listing for ToolFrame dropdown menu. Has CheckIcon (left), custom icon,
        then label (right)
        
        Background color changes when the cursor hovers
        
        User clicks to select/deselect. Selection status is indicated by the
        CheckIcon (left)
    '''
    def __init__(self, master:Frame, label:str, icon_path:str, callback=None,
                 inactive_bg:str='#000000', active_bg:str=None, fg:str='#dddddd',
                 active_fg:str='#ffffff', icon_h:int=24, icon_w:int=24, icon_padx=0,
                 font_name='Segoe UI', font_size=10, right_pad=0, selected=False):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param label: str - label title (text displayed to user)
            :param icon_path: path to png file or numpy array
            :param callback: function(status:bool) - called when label is clicked by user
            :param inactive_bg: str (hex code) - background color when unchecked
            :param active_bg: str (hex code) or None - check background color (if different from inactive_bg)
            :param fg: str (hex code) - foreground color
            :param icon_h: int - icon height
            :param icon_w: int - icon width
            :param selected: bool - initial selection status
            :param font_name: str - label font name
            :param font_size: int - label font size
            :param right_pad: int - number of extra spaces after label for x padding
            :param selected: bool - initial selection status
        '''
        super().__init__(master, bg=inactive_bg)
        self.__callback = callback
        self.__text = label
        self.__hovering = False

        active_bg if active_bg is not None else inactive_bg
        self.__bg_colors = [inactive_bg, active_bg]
        self.__fg_colors = [fg, active_fg]

        self.Check = CheckLabel(self, inactive_bg=inactive_bg, active_bg=active_bg,
                                inactive_fg=fg, active_fg=active_fg, h=icon_h, w=icon_w,
                                selected=selected)
        self.Check.pack(side='left', padx=icon_padx)

        self.Icon = Icon(self, icon_path, bg=inactive_bg, fg=fg,
                         bg2='#00ff00', fg2=active_fg, h=icon_h, w=icon_w, status=True) # status is False for not hovering
        self.Icon.pack(side='left', padx=icon_padx)

        self.Label = Label(self, text=self.__text + ' ' * right_pad, bg=inactive_bg,
                           fg=fg, font=(font_name, font_size))
        self.Label.pack(side='left')

        # Event Bindings
        self.bind('<Enter>', self.__hover_enter)
        self.bind('<Leave>', self.__hover_leave)
        self.bind('<Button-1>', self.__click)
        self.Icon.bind('<Button-1>', self.__click)
        self.Label.bind('<Button-1>', self.__click)
        self.Check.bindall('<Button-1>', self.__click)

    def bindall(self, *args, **kwargs):
        '''binds to all sub widgets'''
        self.bind(*args, **kwargs)
        self.Label.bind(*args, **kwargs)
        self.Icon.bind(*args, **kwargs)
        self.Check.bindall(*args, **kwargs)

    def __hover_enter(self, event=None):
        '''called when cursor hovers on popup'''
        self.__hovering = True
        self.__set_bg(True)

    def __hover_leave(self, event=None):
        '''called when cursor leaves popup'''
        self.__hovering = False
        self.__set_bg(False)

    def __set_bg(self, hovering:bool):
        '''sets background color based on hover status'''
        self.Check.set_bg(hovering)
        self.Icon.set(hovering)
        self.Label.config(bg=self.__bg_colors[hovering], fg=self.__fg_colors[hovering])
        self.config(bg=self.__bg_colors[hovering])

    def __click(self, event=None):
        '''cursor clicks to toggle selection status'''
        self.set(not self.get())
        if self.__callback is not None:
            self.__callback(self.get())

    def set(self, status:bool):
        '''sets selection status'''
        self.Check.set(status)

    def get(self) -> bool:
        '''returns current selection status'''
        return self.Check.get()
    
    def get_label(self) -> str:
        '''returns label text'''
        return self.__text

    def is_hovering(self) -> bool:
        '''returns True if cursor is hovering, otherwise False'''
        return self.__hovering



from tkinter import Frame

from .timer import Timer
from .buttons import PlayerButtons
from .sliders import TimeSlider, PlotScrollBar, DoubleScrollBar

# TODO: make TimeSlider attribute customizable through Player

class Player(Frame):
    ''' Player executes a callback function with precise timing.

        Combines PlayerButtons and a scrollbar to give user control of playback
    
        Includes a scrollbar for user to control playing.
        
        An example use case could be playing frames from a video
    '''
    def __init__(self, master:Frame, callback, delay:float, bg:str='#000000',
                 slider_type:str='single', frame_num=1000, frame_rate=29.97,
                 step_increment:int=150, end_callback=None, start_callback=None,
                 stop_callback=None, skip_callback=None, buttons_on_top=False,
                 buttons_padx_weight=6, simple_slider_width=None,
                 limit_running_callbacks=False):
        ''' Only keeps track of the current frame using Timer

        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param callback: function (step) - called with each iteration
            :param delay: float - seconds between each iteration
                                - would be based on frame rate for video player
            :param bg: str - background color
            :param slider_type: str - options: ['simple', 'single', 'double']
            :param frame_num: int - number of frames
            :param step_increment: float - number of steps to skip forward or
                                         - back when step buttons are clicked
            :param end_callback: function () - called when final step is reached
                                             - not called if looping is on
            :param start_callback: function (current_step) - called when timer is started
            :param stop_callback: function () - called when timer is stopped
            :param skip_callback: function (current_step) - called when step is incremented
            :param buttons_on_top: bool - True to put buttons on top, or False for bottom
            :param buttons_padx_weight: int - weight of padding for PlayerButtons
            :param limit_running_callbacks: bool - if True, does not call callback function immediately when slider is moved while running
                                                 - to avoid calling the callback function in multiple different threads
        '''
        self.__callback = callback
        self.__end_callback = end_callback
        self.__frame_rate = frame_rate
        self.__step_increment = step_increment
        self.__slider_type = slider_type
        self.__limit_running_callbacks = limit_running_callbacks
        super().__init__(master, bg=bg)

        # compute sides for widget packing
        sides = ['bottom', 'top']
        slider_side = sides[not buttons_on_top]
        buttons_side = sides[buttons_on_top]

        self.__Timer = Timer(delay, self.__timer_update, end_callback=self.__end,
                             start_callback=start_callback, stop_callback=stop_callback,
                             skip_callback=skip_callback, min_step=0, max_step=frame_num)

        if slider_type == 'simple':
            self.__Slider = TimeSlider(self, self.__slider_update, bg=bg,
                                       frame_num=frame_num, active_line_color='#ffffff',
                                       active_line_hover_color='#13ce12', slider_color='#ffffff',
                                       hide_slider=True, steps_per_sec=1/delay,
                                       slider_type='circle', width=simple_slider_width)
        elif slider_type == 'single':
            self.__Slider = PlotScrollBar(self, self.__slider_update, None, frame_num,
                                          min_frame=0, start_frame=0, frame_rate=self.__frame_rate,
                                          height=60, bg=bg)
        elif slider_type == 'double':
            self.__Slider = DoubleScrollBar(self, self.__slider_update, None, frame_num,
                                            min_frame=0, start_frame=0, frame_rate=self.__frame_rate,
                                            bg=bg)
        self.__Slider.pack(side=slider_side, fill='x')

        self.__Buttons = PlayerButtons(self, bg, self.__Timer.start, self.__Timer.stop,
                                       self.step_forward, self.step_back,
                                       self.to_end, self.to_start,
                                       padx_weight=buttons_padx_weight)
        self.__Buttons.pack(side=buttons_side, fill='x')

    def start(self):
        '''starts player - called externally - not from buttons'''
        self.__Timer.start()
        self.__Buttons.to_stop()

    def stop(self):
        '''stops player - called externally - not from buttons'''
        self.__Timer.stop()
        self.__Buttons.to_play()

    def __callback_status(self):
        '''returns True if callback should be called (for slider or command events)'''
        return not self.__limit_running_callbacks or not self.is_running()

    def step_forward(self):
        '''same as clicking 'skip forward' button in PlayerButtons'''
        self.__Timer.increment(self.__step_increment, callback=self.__callback_status())

    def step_back(self):
        '''same as clicking 'step back' button in PlayerButtons'''
        self.__Timer.increment(-self.__step_increment, callback=self.__callback_status())

    def to_start(self):
        '''same as clicking 'previous' button in PlayerButtons'''
        self.__Timer.reset(callback=self.__callback_status())

    def to_end(self):
        '''same as clicking 'next' button in PlayerButtons'''
        self.__Timer.to_end(callback=self.__callback_status())

    def set_frame(self, frame:int, callback=True):
        '''updates the current frame and calls callback function (if callback==True)'''
        self.__Timer.set(frame)
        self.__Slider.set_frame(frame)
        if callback:
            self.__callback(frame)

    def set_frame_num(self, frame_num:int):
        '''updates number of steps - pushes to timer and scrollbar'''
        self.__Timer.set_max_step(frame_num)
        self.__Slider.set_frame_num(frame_num, self.__frame_rate)

    def set_frame_rate(self, frame_rate:float):
        '''updates frame rate - number of frames stays sthe same'''
        self.__frame_rate = frame_rate
        self.__Timer.set_delay(1 / self.__frame_rate)
        self.__Slider.set_frame_num(self.__Timer.get_max_step(), self.__frame_rate)

    def set_delay(self, delay:float):
        '''updates delay of timer
        intended for when Player is not being used to play video frames'''
        self.__Timer.set_delay(delay)
        if self.__slider_type == 'simple':
            self.__Slider.set_steps_per_sec(1 / delay)

    def set_increment(self, inc:int):
        '''sets number of steps to skip when 'skip forward' and 'skip back'
        buttons are clicked'''
        self.__step_increment = inc

    def get_step(self) -> int:
        '''returns the current step'''
        return self.__Timer.get_step()

    def get_delay(self) -> float:
        '''returns delay between callbacks (seconds)'''
        return self.__Timer.get_delay()
    
    def get_max_step(self) -> int:
        '''returns the maximum number of steps'''
        return self.__Timer.get_max_step()

    def get_frame_rate(self) -> float:
        '''returns frame rate'''
        return self.__frame_rate

    def is_running(self) -> bool:
        '''returns True if Player is running, otherwise False'''
        return self.__Timer.is_running()

    def __timer_update(self, step:int):
        '''called by timer when playing'''
        self.__Slider.set_frame(step)
        self.__callback(step)

    def __slider_update(self, step:int):
        '''called when slider is moved by user
        it does not matter whether timer is running or not'''
        self.__Timer.set(step)
        if self.__callback_status():
            self.__callback(step)

    def __end(self):
        '''called when player reaches the end (max step)'''
        if self.__Buttons.is_looped():
            self.__Timer.reset()
            self.__Timer.start()
        else: # no loop - actually stopping
            self.__Buttons.to_play()
            if self.__end_callback is not None:
                self.__end_callback()
    



__version__ = "0.0.6"

# Import widgets
from .aspect_frame import *
from .buttons import *
from .collapse_frame import CollapseFrame
from .dropdowns import *
from .entry_boxes import *
from .file_dialog import *
from .function_progress import *
from .icon_labels import *
from .label_dropdown import *
from .labels import *
from .pdf_display import PdfDisplay
from .player import *
from .progress_bar import *
from .scrollable_frame import *
from .sliders import *
from .temp_label import *
from .temp_menu import *
from .text_boxes import *
from .timer import *
from .tool_frame import *
from .tool_tip import *

# not importing from .canvas_items because these are not intended
# to be used outside of chichitk

# not importing from .icons because this just contains icons as arrays


import time
from threading import Thread


class Timer:
    ''' Performs precisely timed callbacks
    
        The timer is given a set number of steps which could be frames in a
        video, seconds until an alarm, etc. The timer will count up, calling a
        callback function with each iteration, until the set number of steps is
        reached.

        The callback functions must accept a single integer argument, which is
        the current step.

        The timer works by storing the current time when it is started and
        computing the step of each iteration based on the time elasped since
        the timer started. It is not guaranteed that the callback function will
        be called with exactly sequential steps since step(s) will be skipped
        if the callback function takes longer than expected to execute.

        The number of steps, and the delay between steps can be changed at any
        time (even while the timer is running)
        
        The simplest usage is a stopwatch, where delay could be 1 for seconds or
        0.001 for miliseconds. The callback will be called to increment the time
    '''
    
    def __init__(self, delay:float, callback, end_callback=None,
                 start_callback=None, stop_callback=None, skip_callback=None,
                 min_step=0, max_step=None, reset_end_start=True,
                 track_execution_time=False):
        '''sets timer with number of steps and delay, both of which can be changed later
        
        Parameters
        ----------
            :param delay: float - time to elapse between steps
            :param callback: function (current_step) - called with each step
            :param end_callback: function () - called when final step is reached
            :param start_callback: function (current_step) - called when timer is started
            :param stop_callback: function () - called when timer is stopped
            :param skip_callback: function (current_step) - called when step is incremented
            :param min_step: int - lowest allowed step (default=0)
            :param max_step: int - number of steps in timer (default=None - unlimited)
            :param reset_end_start: bool - If True, resets timer when starting from the end
            :param track_execution_time: bool - only use when debugging (more computationally expensive)

        Note
        ----
            * Start, stop, skip, and end callbacks are always called regardless of the
                manner in which the event occurs
            * For example, if you call the Timer.start() method, the start
                callback will be called
            * However, if Timer.start() is called when timer is already running,
                the start callback will not be called
            * Timer.to_end() does not call the skip function since this stops the
                timer
        '''
        self.__delay = delay
        self.__callback = callback
        self.__end_callback, self.__skip_callback = end_callback, skip_callback
        self.__start_callback, self.__stop_callback = start_callback, stop_callback
        self.__min_step, self.__max_step = min_step, max_step
        self.__current_step = min_step
        self.__start_step = min_step # relevant while running
        self.__track_execution_time = track_execution_time
        self.__reset_end_start = reset_end_start
        self.__running = False

    def start(self):
        '''
        Purpose
        -------
            starts timer at current step
            if you need to start at a specific step, call set() first then start()

        Returns
        -------
            :return: bool - False if timer was already running, otherwise True
        '''
        if not self.__running: # cannot start timer if already started
            # if starting from the end, so go back to start
            if self.at_end() and self.__reset_end_start:
                self.__current_step = self.__min_step
            self.__running = True
            self.__start_time = time.time()
            self.__last_time = self.__start_time
            self.__start_step = self.__current_step
            Thread(target=self.__run).start()
            if self.__start_callback is not None:
                self.__start_callback(self.__current_step)
            return True
        return False

    def stop(self):
        '''
        Purpose
        -------
            stops timer immediatly - next callback will be halted
            
        Returns
        -------
            :return: bool - False if timer was already stopped, otherwise True
        '''
        if self.__running:
            self.__running = False
            if self.__stop_callback is not None:
                self.__stop_callback()
            return True
        return False

    def reset(self, callback=True):
        '''resets step to min_step and calls callback function
        does not stop the timer'''
        self.increment(self.__min_step - self.__current_step, callback=callback)

    def to_end(self, callback=True):
        '''sets step to max_step, stops timer, and calls callback function'''
        assert self.__max_step is not None, 'Skipped timer to end when max_step is not defined'
        self.stop()
        if self.__end_callback is not None:
            self.__end_callback()
        self.__current_step = self.__max_step
        if callback:
            self.__callback(self.__current_step)

    def set(self, step:int):
        '''
        Purpose:
            sets timer to the current step.
            if timer is currently running, it will continue uninterrupted
            the next callback will reflect the updated step
        Pre-conditions:
            :param step: int - replaces current step
        '''
        assert isinstance(step, int), 'Step must be an integer'
        if step < self.__min_step or (self.__max_step is not None and step > self.__max_step):
            raise ValueError(f"Step is out of range: {step}.")
        self.__start_step += step - self.__current_step # for running computation
        self.__current_step = step
        if self.__skip_callback is not None:
            self.__skip_callback(self.__current_step)

    def increment(self, inc:int, callback=False):
        '''changes current step by the given increment (can be positive or negative)
        forces current step to stay between min_step and max_step
        only calls callback function if callback == True and timer is running
        '''
        inc = max(self.__min_step - self.__current_step, inc) # enforce upper limit
        if self.__max_step is not None: # enforce upper limit
            inc = min(self.__max_step - self.__current_step, inc)
        self.__current_step += inc
        self.__start_step += inc # so that running step is computed correctly
        if callback and not self.__running:
            self.__callback(self.__current_step)
        if self.__skip_callback is not None:
            self.__skip_callback(self.__current_step)

    def get_step(self) -> int:
        '''returns the current step'''
        return self.__current_step

    def set_delay(self, delay:float):
        '''
        Purpose:
            updates the delay between callbacks.
            if timer is currently running, the next callback will occur
            at the previously scheduled time, and callbacks after that will
            reflect the updated delay
        Pre-conditions:
            :param delay: float - new delay between callbacks
        '''
        self.__delay = delay

    def get_delay(self) -> float:
        '''returns delay between callbacks (seconds)'''
        return self.__delay

    def set_max_step(self, max_step:int):
        '''updates max step'''
        self.__max_step = max_step

    def get_max_step(self):
        '''returns max_step - must not call this is max_step is undefined'''
        assert self.__max_step is not None, 'Tried to get max_step when it is undefined'
        return self.__max_step
    
    def at_end(self):
        '''returns True if timer is at the end, otherwise False'''
        return self.__max_step is not None and self.__current_step >= self.__max_step
    
    def is_running(self):
        '''returns True if timer is running, otherwise False'''
        return self.__running

    def __run(self):
        '''private method to actually run timer
        private because it must always be called in a Thread
        '''
        while self.__running:
            # compute time to wait before next iteration - approx
            current_time = time.time()
            elapsed_time = current_time - self.__last_time # since last iteration
            sleep_time = max(0, self.__delay - elapsed_time)
            self.__last_time = current_time

            # compute step based on time since start
            step_inc = (self.__last_time - self.__start_time) / self.__delay
            self.__current_step = int(round(self.__start_step + step_inc, 0))

            # check if timer has reached the end
            if self.__max_step is not None and self.__current_step >= self.__max_step:
                self.stop()
                if self.__end_callback is not None:
                    self.__end_callback()
                break

            self.__callback(self.__current_step)

            if self.__track_execution_time:
                execution_time = time.time() - self.__last_time
                if execution_time > self.__delay:
                    print(f'Callback took {execution_time:.4f}s, which is longer than delay: {self.__delay:.4f}s')

            time.sleep(sleep_time) # wait before next iteration




from tkinter import Frame, Label, filedialog

import os

from .buttons import IconButton
from .icons import icons


class FileDialog(Frame):
    ''' Widget that allows user to load and store file names

        The currently loaded file can be retrieved using the .get() method
    '''
    def __init__(self, master, label:str, file_types:list, bg:str,
                 fg:str='#ffffff', file_active_fg:str='#ffffff',
                 file_inactive_fg:str='#ffffff', inactive_bg:str='#000000',
                 loaded_bg:str='#00ff00', load_callback=None,
                 label_font_name='Segoe UI bold', file_font_name='Segoe UI',
                 label_font_size=10, file_font_size=10,
                 target_filenames:list=None,
                 row=None, start_col=0, padx=0, pady=0):
        '''frame to load and access filenames
        specify row to grid components directly in master
        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param label: str - label to indicate what the file is for
            :param file_types: list[str] - list of filetypes such as 'wav' or 'csv'
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - label foreground color
            :param file_active_fg: str (hex code) - filename foreground when loaded
            :param file_inactive_fg: str (hex code) - filename foreground when not loaded
            :param inactive_bg: str (hex code) - filename background when not loaded
            :param loaded_bg: str (hex code) - filename background when loaded
        file_type : list of str - file types such as 'wav' or 'png'
        load_callback : 1 argument function (filename) - called when a new file is loaded
        '''
        Frame.__init__(self, master, bg=bg)
        self.__start_folder = None
        self.__load_callback = load_callback
        self.__filename = None
        self.__target_filenames = target_filenames if target_filenames else []
        self.inactive_bg, self.loaded_bg = inactive_bg, loaded_bg
        self.file_active_fg, self.file_inactive_fg = file_active_fg, file_inactive_fg
        self.__filetypes = [(f'{t.upper()} Files', f'*.{t}') for t in file_types] + [('All Files', '*.*')]

        if row != None:
            file_label = Label(master, text=label, bg=bg, fg=fg,
                               font=(label_font_name, label_font_size))
            file_label.grid(row=row, column=start_col, padx=padx, pady=pady, sticky='nsew')
            load_button = IconButton(master, icons['file_upload'], self.browse_file,
                                     popup_label='Upload File', selectable=False,
                                     bar_height=0, inactive_bg=bg)
            load_button.grid(row=row, column=start_col + 1, padx=padx, pady=pady, sticky='nsew')
            self.file_label = Label(master, text='No File Loaded', bg=inactive_bg,
                                    fg=self.file_inactive_fg,
                                    font=(file_font_name, file_font_size))
            self.file_label.grid(row=row, column=start_col + 2, padx=padx, pady=pady, sticky='nsew')
        else:
            file_label = Label(self, text=label, bg=bg, fg=fg,
                               font=(label_font_name, label_font_size))
            file_label.pack(side='left', fill='x', expand=True)
            load_button = IconButton(self, icons['file_upload'], self.browse_file,
                                     popup_label='Upload File', selectable=False,
                                     bar_height=0, inactive_bg=bg)
            load_button.pack(side='right')
            self.file_label = Label(self, text='No File Loaded', bg=inactive_bg,
                                    fg=self.file_inactive_fg,
                                    font=(file_font_name, file_font_size))
            self.file_label.pack(side='left', fill='x', expand=True)

    def load(self, folder):
        '''sets start folder for browsing files
        
        folder : str - path to song folder such as "C:\\Users\\samue\\Electric Praise Songs\\Tester - Test Song"
        '''
        self.__start_folder = folder
        for filename in self.__target_filenames:
            if os.path.exists(self.__start_folder + '\\' + filename):
                self.load_file(self.__start_folder + '\\' + filename)
                break

    def load_file(self, filename:str):
        '''load file internal command instead of user browsing file'''
        if isinstance(filename, str) and filename != '' and os.path.exists(filename): # didnt click 'cancel' instead of selecting a file
            self.__filename = filename
            if '/' in self.__filename:
                text = self.__filename.split('/')[-1]
            elif '\\' in self.__filename:
                text = self.__filename.split('\\')[-1]
            else:
                text = self.__filename
            self.file_label.config(text=text, bg=self.loaded_bg,
                                   fg=self.file_active_fg)
            if self.__load_callback:
                self.__load_callback(self.__filename)

    def browse_file(self):
        '''user click upload button to browse file from computer'''
        self.load_file(filedialog.askopenfilename(initialdir=self.__start_folder,
                                                  title='Select File', filetypes=self.__filetypes))

    def is_loaded(self):
        '''returns True if a file is loaded, otherwise False'''
        return self.__filename is not None

    def get(self):
        '''returns filename assuming a file has been loaded
        returns None if no file is loaded'''
        return self.__filename

    def remove(self):
        '''removes and discards the file that is currently loaded'''
        self.__filename = None
        self.file_label.config(text='No File Loaded', bg=self.inactive_bg,
                               fg=self.file_inactive_fg)



from tkinter import Frame, Text


def consecutive_spaces(text:str, consec_char:str=' ', leading:str=True,
                       trailing:str=True):
    '''checks if consec_char occurs consecutively in text
    
    Parameters
    ----------
        :param text: str - must not contain newline characters - '\n'
        :param consec_char: str - single character to check for text
        :param leading: bool - if True, checks if first character is consec_char
        :param trailing: bool - if True, checks if last character is consec_char

    Returns:
        :return: list[tuple[int]] - indices where consecutive consec_char occurs (upper bound exclusive)
                                  - or None if there are no consecutive instances of consec_char
    '''
    indices = []
    start = 0
    end = None
    for i, char in enumerate(text):
        if char == consec_char: # start new sequence or extend current
            if start == None:
                start = i
            else:
                end = i
        elif start != None and end != None: # end last sequence
            indices.append((start, end + 1))
            start = None
            end = None
        elif start != None: # cancel last sequence - only one instance of consec_char
            if i == 1 and leading: # first character is consec_char
                indices.append((0, 1))
            start = None
    if trailing and start != None and end != None: # trailing white space
        indices.append((start, end + 1))
    if start != None:
        indices.append((start, start + 1))
    if len(indices) == 0:
        return False
    return indices


class TextBox(Frame):
    ''' Text widget that counts the lines just like a code editor and can
        call a callback function whenever the text is edited by the user
        
        TextBox can optionally display an error color when there are blank
        lines or consecutive spaces
    '''
    def __init__(self, master, callback=None, bg:str='#ffffff', fg:str='#000000',
                 cursor_color=None, disabled_bg=None, disabled_fg=None,
                 error_bg:str='#3c2525', error_highlight_bg:str='#ff0000',
                 error_highlight_fg:str='#000000', track_fg:str='#a6a6a6',
                 track_active_fg:str='#ffffff', active_line_indices:list=None,
                 font_name:str='Consolas', font_size:int=15, wrap='none',
                 focus_in_function=None, focus_out_function=None,
                 scroll_callback=None, line_num_callback=None,
                 check_blank_lines:bool=True, check_consecutive_spaces:bool=True,
                 line_numbers_labels=True, width=None, height=None,
                 justify='left', inactive_justify='center'):
        '''Text box with number lines and callback when text box is edited
        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param callback: function (str) - called whenever text box is modified
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - main text color
            :param cursor_color: str (hex code) - cursor color - if different from fg
            :param disabled_bg: str (hex code) - background color when disabled
            :param disabled_fg: str (hex code) - foreground color when disabled
            :param error_bg: str (hex code) - background color when there is an error
            :param error_highlight_bg: str (hex code) - highlight on text causing error
            :param error_highlight_fg: str (hex code) - color of text cauing error
            :param track_fg: str (hex code) - color of track text - line numbers
            :param track_active_fg: str (hex code) - color of active lines in track box
            :param active_line_indices: list[int] - indices of active lines
            :param font_name: str - font for text box and line numbers
            :param font_size: int - font size for text box and line numbers
            :param wrap: str Literal['none', 'char', 'word'] - wrap setting
            :param focus_in_function: function () - called when text box takes focus
            :param focus_out_function: function () - called when text box loses focus
            :param scroll_callback: function(yview_moveto) - called when text box is scrolled
            :param line_num_callback: function(line_num) - called when number of lines is changed
            :param check_blank_lines: bool - if True, show error when there are blank lines
            :param check_consecutive_spaces: bool - if True, show error when there are consecutive spaces
            :param line_numbers_labels: bool - if True, show line numbers
        '''
        self.callback_function = callback
        self.scroll_callback = scroll_callback
        self.line_num_callback = line_num_callback
        self.line_num = 0
        self.active_line_indices = active_line_indices if active_line_indices is not None else []
        self.bg, self.fg = bg, fg
        self.track_active_fg = track_active_fg
        self.error_bg = error_bg
        self.error_highlight_bg = error_highlight_bg
        self.error_highlight_fg = error_highlight_fg
        self.disabled_bg = disabled_bg if disabled_bg else bg
        self.disabled_fg = disabled_fg if disabled_fg else fg
        self.justify, self.inactive_justify = justify, inactive_justify
        Frame.__init__(self, master, bg=bg)
        self.good_format = True # False if there are errors in the text box
        self.check_blank_lines = check_blank_lines
        self.check_consecutive_spaces = check_consecutive_spaces
        cursor_color = cursor_color if cursor_color else fg

        self.track = Text(self, width=4, height=height, font=(font_name, font_size),
                          bg=bg, fg=track_fg, wrap='none', bd=0)
        if line_numbers_labels:
            self.track.pack(side='left', fill='y')
        self.track.insert('end', '  1')
        self.track.config(state='disabled')
        # undo must be False because Ctrl+z causes infinite loop (no idea why)
        self.box = Text(self, width=width, height=height, font=(font_name, font_size),
                        bg=bg, fg=fg, insertbackground=cursor_color, undo=False,
                        wrap=wrap, yscrollcommand=self.__box_scroll, bd=0)
        self.box.pack(side='right', fill='both', expand=True)
        self.box.tag_add("justify", 1.0, "end")
        self.box.tag_configure('justify', justify=self.justify)
        self.box.bind("<<TextModified>>", self.callback)
        # must not set yscrollcommand until after self.box is defined
        # to avoid AttributeError
        self.track.config(yscrollcommand=self.__track_scroll)
        #self.box.bind("<Key>", self.callback)
        if focus_in_function is not None:
            self.box.bind("<FocusIn>", focus_in_function)
        if focus_out_function is not None:
            self.box.bind("<FocusOut>", focus_out_function)

        # create a proxy for the underlying widget
        self.box._orig = self.box._w + "_orig"
        self.box.tk.call("rename", self.box._w, self.box._orig)
        self.box.tk.createcommand(self.box._w, self._proxy)

    def __box_scroll(self, a, b):
        '''called when text box is scrolled with mousewheel'''
        self.track.yview_moveto(a)
        if self.scroll_callback is not None:
            self.scroll_callback(a)

    def __track_scroll(self, a, b):
        '''called when track box is scrolled with mousewheel'''
        self.box.yview_moveto(a)
        if self.scroll_callback is not None:
            self.scroll_callback(a)

    def _proxy(self, command, *args):
        '''facilitates callback - called whenever Text box is edited by user'''
        cmd = (self.box._orig, command) + args
        try:
            result = self.box.tk.call(cmd)
        except:
            # As far as I can tell, this error only occurs when pasting
            # pasting involves: 1. deleting selected text 2. pasting copied text
            # tk is trying to delete selected text when no text is selected
            # Pasting worked fine when there was text selected to "overwrite"
            # Because tk is trying to delete "nothing", this error an be skipped without effect
            # Hopefully this doesn't bite me in the ass down the road...
            # Many frustrating hours were spent trying to solve this... :/
            return None
        if command in ("insert", "delete", "replace"):
            self.box.event_generate("<<TextModified>>")
        return result

    def callback(self, event=None):
        '''called whenever text box is modified'''
        text = self.box.get('1.0', 'end-1c')
        y_pos = self.box.yview()[0]
        self.track.config(state='normal')
        self.track.delete(0.0, 'end')
        self.track.insert('end', '\n'.join([f'{x} ' for x in range(1, text.count('\n') + 2)]))
        self.track.config(state='disabled')
        self.track.tag_delete('right')
        self.track.tag_add("right", 1.0, "end")
        self.track.tag_configure("right", justify='right')
        self.set_active_lines(self.active_line_indices) # reset after updating track box
        if text.count('\n') + 1 != self.line_num: # number of lines has changed
            self.line_num = text.count('\n') + 1
            if self.line_num_callback is not None:
                self.line_num_callback(self.line_num)

        # do basic error checking
        self.good_format = True
        self.box.tag_delete('empty_line')
        self.box.tag_delete('consecutive_space')
        if text != '' and (self.check_blank_lines or self.check_consecutive_spaces):
            for i, line in enumerate(text.split('\n')):
                if self.check_blank_lines and line == '':
                    self.good_format = False
                    self.box.tag_add('empty_line', f'{i + 1}.0', f'{i + 2}.0')
                    continue
                # check for two or more consecutive spaces
                if self.check_consecutive_spaces and consecutive_spaces(line, trailing=False) != False:
                    self.good_format = False
                    for tup in consecutive_spaces(line, trailing=False):
                        self.box.tag_add('consecutive_space',
                                         f'{i + 1}.{tup[0]}',
                                         f'{i + 1}.{tup[1]}')
            self.box.tag_config('empty_line', background=self.error_highlight_bg,
                                foreground=self.error_highlight_fg)
            self.box.tag_config('consecutive_space', background=self.error_highlight_bg,
                                foreground=self.error_highlight_fg)
            if self.good_format:
                self.box.config(bg=self.bg)
                self.track.config(bg=self.bg)
            else:
                self.box.config(bg=self.error_bg)
                self.track.config(bg=self.error_bg)

        self.box.yview_moveto(y_pos)
        self.track.yview_moveto(y_pos)
        if self.callback_function:
            self.callback_function(text)

    def set_active_lines(self, line_indices:list):
        '''changes the foreground in track box for the given line indices
        :param line_indices: list of ints - indices of lines to highlight
        '''
        self.active_line_indices = line_indices
        self.track.tag_delete('active_track_line')
        for i in line_indices:
            self.track.tag_add('active_track_line', f'{i + 1}.0', f'{i + 2}.0')
        self.track.tag_config('active_track_line', foreground=self.track_active_fg)

    def get(self, strip=True) -> str:
        '''returns entire text in text box

        Parameters
        -----------
            :param strip: bool - if True, removes spaces and newline characters for beginning and end
        '''
        if strip:
            return self.box.get(0.0, 'end').strip()
        return self.box.get(0.0, 'end')

    def clear(self):
        '''clears all text from text box'''
        self.box.delete(0.0, 'end')

    def insert(self, text:str):
        '''inserts text at the end of text box'''
        self.box.insert('end', text)

    def clear_insert(self, text:str):
        '''clears all text from text box and adds text'''
        self.clear()
        self.insert(text)

    def set_active(self):
        '''
        Purpose:
            sets text box state to 'normal' so that it is interactable
        '''
        self.box.config(state='normal', bg=self.bg, fg=self.fg)
        self.box.tag_add("justify", 1.0, "end")
        self.box.tag_configure('justify', justify=self.justify)
        self.track.config(bg=self.bg)

    def set_inactive(self):
        '''
        Purpose:
            sets text box state to 'disabled' so that it is not interactable
        '''
        self.box.config(state='disabled', bg=self.disabled_bg, fg=self.disabled_fg)
        self.box.tag_add("justify", 1.0, "end")
        self.box.tag_configure('justify', justify=self.inactive_justify)
        self.track.config(bg=self.disabled_bg)




from tkinter import Toplevel, Widget

from .icon_labels import IconCheckLabel


class ToolFrame(Toplevel):
    ''' Popup to display list of selectable items
    
        Should be connected to a Button to appear when button is clicked
    '''
    def __init__(self, master:Widget, parameters:list, callback=None,
                 bg:str='#000000', active_bg:str='#444444',
                 fg:str='999999', active_fg:str='#ffffff',
                 label_padx=1, label_pady=2, right_pad=2):
        '''
        Parameters
        ----------
            :param master: Widget - parent widget (should be Button)
            :param parameters: list[dict] - each dict has keys: ['label', 'icon']
            :param callback: function(index:int, status:bool) - called when a label is clicked
            :param right_pad: int - number of extra spaces after each label for x padding
        '''
        super().__init__(master, bg=bg)
        self.__callback = callback
        self.__visible = False # True when the popup window is visible

        # make window invisible, on the top, and strip all window decorations/features
        # change alpha to make window visible
        self.attributes('-alpha', 0, '-topmost', True)
        self.overrideredirect(1)

        # add labels
        self.__labels: list[IconCheckLabel] = []
        for i, d in enumerate(parameters):
            selected = d['selected'] if 'selected' in d.keys() else False
            label = IconCheckLabel(self, d['label'], d['icon'],
                                   callback=lambda s, x=i: self.__click_callback(x, s),
                                   inactive_bg=bg, active_bg=active_bg,
                                   fg=fg, active_fg=active_fg, right_pad=right_pad,
                                   selected=selected, icon_padx=0)
            label.pack(side='top', fill='x', padx=label_padx, pady=label_pady)
            self.__labels.append(label)

    def bindall(self, *args, **kwargs):
        '''binds to all sub widgets'''
        self.bind(*args, **kwargs)
        for label in self.__labels:
            label.bindall(*args, **kwargs)

    def __click_callback(self, index:int, status:bool):
        '''called when a label is clicked by user'''
        if self.__callback is not None:
            self.__callback(index, status)

    def show(self, event=None):
        '''called when parent button is clicked
        compute coordinates based on master (button) and set alpha to visible
        '''
        # check parent position (button that calls tool_tip popup)
        widget_width = self.master.winfo_width()
        widget_height = self.master.winfo_height()
        widget_x = self.master.winfo_rootx()
        widget_y = self.master.winfo_rooty()

        # compute popup geometry - may need to use self.frame
        w = self.winfo_width()
        h = self.winfo_height()
        x = widget_x # align to left edge of parent
        y = widget_y + widget_height # directly below widget

        # check root coordinates (main app window)
        root_width = self.master.winfo_toplevel().winfo_width()
        root_height = self.master.winfo_toplevel().winfo_height()
        root_x = self.master.winfo_toplevel().winfo_rootx() # x position of app on computer screen
        root_y = self.master.winfo_toplevel().winfo_rooty() # y position of app on computer screen

        # adjust so that popup is entirely within the root window
        if x + w > root_x + root_width: # enforce right edge
            x = root_x + root_width - w
        if y + h > root_y + root_height: # enforce bottom edge
            y = root_y + root_height - h
        x = max(root_x, x) # enforce left edge
        y = max(root_y, y) # enforce top edge

        # ensure tool_tip is not on top of widget - this makes a button unclickable!
        # this can only happen if tool_tip was moved up by bottom enforcement, so move it to above widget
        if y < widget_y + widget_height:
            y = widget_y - h # position tool_tip above widget

        self.geometry(f'{w}x{h}+{x}+{y}') # apply geometry
        self.attributes('-alpha', 1) # make window visible
        self.__visible = True

    def hide(self, event=None):
        '''should be called when cursor leaves popup window and parent button
        makes popup invisible
        '''
        self.attributes('-alpha', 0)
        self.__visible = False

    def select_all(self):
        '''selects all labels'''
        for label in self.__labels:
            label.set(True)
    
    def deselect_all(self):
        '''deselects all labels'''
        for label in self.__labels:
            label.set(False)

    def set(self, d:dict):
        '''sets selection status of labels
        :param d: dict - label(str):status(bool) for each item
        '''
        for label, value in d.items():
            for l in self.__labels:
                if label == l.get_label():
                    l.set(value)
                    break

    def get(self) -> dict:
        '''returns dictionary with label(str):status(bool) for each item'''
        return {d.get_label():d.get() for d in self.__labels}

    def is_visible(self) -> bool:
        '''returns True if window is visible, otherwise False'''
        return self.__visible

    def contains_point(self, x:int, y:int) -> bool:
        '''returns True if the given (x, y) point is in this Toplevel window, otherwise False'''
        x0 = self.winfo_rootx()
        y0 = self.winfo_rooty()
        x1 = x0 + self.winfo_width()
        y1 = y0 + self.winfo_height()
        # only y0 is >= so that popup does not disappear when moving from button to popup
        return x > x0 and x < x1 and y >= y0 and y < y1



from tkinter import Entry, StringVar


class CheckEntry(Entry):
    ''' Extension of tk.Entry to check text as it is entered and only allow
        certain characters to be entered
    '''
    def __init__(self, master, default='', allowed_chars=None, max_len=None,
                 check_function=None, exit_function=None, justify='left',
                 bg:str='#ffffff', fg:str='#000000', disabled_bg=None,
                 disabled_fg=None, error_color='#ff0000', width=0,
                 font_name='Segoe UI', font_size=10, editable=True,
                 hide_char=None, select_first=False, focus_first=False,
                 entry_on_function=None, entry_off_function=None, **kwargs):
        '''entry box to check text as it is entered
        
        Parameters
        ----------
            :param master: tk.Frame
            :param default: str - text automatically inserted into entry box
            :param allowed_chars: str or list of str - characters the can be entered
            :param max_len: int - maximum number of characters in box
            :param check_function: function (str) -> bool - text that fails check_function will trigger error_color
            :param exit_function: function (str) - optionally called when focus leaves
            :param justify: Literal['left', 'right', 'center'] - text position
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - foreground color
            :param disabled_bg: str (hex code) - optionally different from bg
            :param disabled_fg: str (hex code) - optionally different from fg
            :param error_colors: str (hex code) - background color when there is an error
            :param width: int - default width of text box
            :param editable: bool - if False, entry box will be disabled (uninteractable)
            :param hide_char: str or None - character to display in entry box instead of text
                                          - example: '*' for password entry
            :param select_first: bool - if True, will select text upon initiation
            :param focus_first: bool - if True will focus CheckEntry upon initiation
            :param entry_on_function: function (event) - called with focus on
            :param entry_off_function: function (event) - called with focus off
        '''
        self.allowed_chars = allowed_chars
        self.max_len = max_len
        self.__check_function = check_function
        self.__exit_function = exit_function
        self.bg = bg
        self.error_color = error_color
        self.just_started = True
        disabled_bg = disabled_bg if disabled_bg else bg
        disabled_fg = disabled_fg if disabled_fg else fg

        self.sv = StringVar()
        self.sv.trace("w", lambda name, index, mode: self.entry_callback())
        Entry.__init__(self, master, width=width, font=(font_name, font_size),
                       textvariable=self.sv, fg=fg, bg=bg, insertbackground=fg,
                       disabledbackground=disabled_bg, disabledforeground=disabled_fg,
                       justify=justify, highlightthickness=0, borderwidth=0,
                       state='normal' if editable else 'disabled', show=hide_char,
                       **kwargs)
        if entry_on_function is not None:
            self.bind('<FocusIn>', entry_on_function)
        if entry_off_function is not None:
            self.bind('<FocusOut>', entry_off_function)
        if self.__exit_function is not None:
            for tag in ['<Return>', '<Tab>', '<FocusOut>']:
                self.bind(tag, self.exit)
        self.activate(text=default, select=select_first, focus=focus_first)

    def activate(self, text=None, select=True, focus=True):
        '''takes focus, replaces current text with text, and selects all text'''
        if text != None:
            self.delete(0, 'end')
            self.insert('end', text)
        if focus:
            self.focus_set()
        if select:
            self.select_range(0, 'end')

    def entry_callback(self):
        '''called whenever text in Entry is modified'''
        if self.just_started: # so that callback will not be called upon initialization
            self.just_started = False
            return None
        text = self.sv.get()
        self.config(bg=self.bg)
        if text == '':
            if self.__check_function is not None: # this is only here because of the misuse of CheckEntry as a search bar
                self.__check_function(text)
            return None
        if (self.allowed_chars is not None and not min([c in self.allowed_chars for c in text])) or (self.max_len and len(text) > self.max_len):
            cursor_index = self.index('insert')
            # remove character that was just entered
            self.delete(str(cursor_index - 1), str(cursor_index))
            text = super().get() # update text to current text in entry box
        if self.__check_function is not None and not self.__check_function(text):
            self.config(bg=self.error_color)

    def set_allowed_chars(self, allowed_chars:str):
        '''updates allowed characters in entry box
        :param allowed_chars: str | list[str] | None
        '''
        self.allowed_chars = allowed_chars

    def set_bg(self, bg:str):
        '''updates background color'''
        self.bg = bg
        self.config(bg=self.bg)

    def get(self):
        '''returns text currently in CheckEntry'''
        return self.sv.get()
    
    def clear(self):
        '''
        Purpose:
            removes all text from entry box
        Pre-conditions:
            (none)
        Post-conditions:
            removes all text from entry box
        Returns:
            (none)
        '''
        self.activate(text='', select=False, focus=False)

    def exit(self, event=None):
        '''calls exit function and destroys widget'''
        if self.__exit_function is not None:
            self.__exit_function(self.get())
        self.destroy()

class ColorEntry(CheckEntry):
    ''' Special version of CheckEntry that allows user to enter a hex code.
        As such, the text must begin with '#', must be 7 characters long, and
        can only contain the characters '0123456789abcdef'
    
        Includes popup that lets user select a color for a color gradient
    '''
    def __init__(self, master, default='#ffffff', callback=None, exit_function=None,
                 justify='left', bg:str='#ffffff', fg:str='#000000',
                 disabled_bg=None, disabled_fg=None, error_color='#ff0000',
                 width=0,  font_name='Segoe UI', font_size=10, editable=True,
                 hide_char=None, select_first=False, focus_first=False, **kwargs):
        '''CheckEntry box to only accept hex codes
        callback function will only be called with valid hex codes
        
        Parameters
        ----------
            :param master: tk.Frame
            :param default: str - text automatically inserted into entry box
            :param callback: function (str) - called when new color is set
            :param exit_function: function (str) - optionally called when focus leaves
            :param justify: Literal['left', 'right', 'center'] - text position
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - foreground color
            :param disabled_bg: str (hex code) - optionally different from bg
            :param disabled_fg: str (hex code) - optionally different from fg
            :param error_colors: str (hex code) - background color when there is an error
            :param width: int - default width of text box
            :param editable: bool - if False, entry box will be disabled (uninteractable)
            :param hide_char: str or None - character to display in entry box instead of text
                                          - example: '*' for password entry
            :param select_first: bool - if True, will select text upon initiation
            :param focus_first: bool - if True will focus CheckEntry upon initiation
        '''
        self.__callback = callback
        CheckEntry.__init__(self, master, default=default,
                            allowed_chars='#0123456789abcdef', max_len=7,
                            check_function=self.color_check_function, exit_function=exit_function,
                            justify=justify, bg=bg, fg=fg, disabled_bg=disabled_bg,
                            disabled_fg=disabled_fg, error_color=error_color,
                            width=width, font_name=font_name, font_size=font_size,
                            editable=editable, hide_char=hide_char,
                            select_first=select_first, focus_first=focus_first,
                            **kwargs)
        
    def color_check_function(self, text:str):
        '''
        Purpose:
            check function for color entry box
            only calls __update_function if color if formatted correctly
        Pre-conditions:
            :param text: str - text for color entry box
        Post-conditions:
            (none)
        Returns:
            :return: bool - True if color text is formatted properly, otherwise False
        '''
        if len(text) == 7 and text[0] == '#' and text.count('#') == 1:
            self.__callback(text)
            return True
        return False




from tkinter import Label, Frame

import time
import numpy as np

from .entry_boxes import CheckEntry
from .buttons import IconButton
from .icons import icons


class EditLabel(Frame):
    ''' Label that changes to an CheckEntry box when user double clicks to
        allow text to be edited
        
        Uses CheckEntry so that text can be checked as it is entered and
        certain characters can be forbidden
    '''
    def __init__(self, master:Frame, text:str, bg:str='#ffffff', fg:str='#000000',
                 hover_bg:str='#cccccc', error_color='#ff0000', callback=None,
                 allowed_chars=None, max_len=None, check_function=None, display_function=None,
                 editable=True, justify='left', focus_out_bind=True, hover_bind=True,
                 hover_enter_function=None, hover_leave_function=None,
                 entry_on_function=None, entry_off_function=None,
                 font_name='Segoe UI', font_size=10, width=0):
        '''entry box to check text as it is entered
        
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param text: str - text displayed on label
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - foreground color
            :param hover_bg: str (hex code) - background color when cursor is hovering
            :param error_color: str (hex code) - background color when there is an error
            :param callback: function (str) - called when label is edited by user
            :param allowed_chars: str or list of str - characters the can be entered
            :param max_len: int - maximum number of characters in box
            :param check_function: function(str) -> bool - called as text is entered
                                                          - if False, changes to error color
            :param display_function: function(str) -> str - alternate text display for label
                            - does not affect text in entry field or text returned by get()
            :param hover_enter_function: function () - called when cursor enters label
            :param hover_leave_function: function () - called when cursor leaves label
            :param entry_on_function: function () - called with self.to_entry
            :param entry_off_function: function () - called with self.to_label
            :param width: int - width of Entry box
        '''
        self.text = text
        self.bg = bg
        self.hover_bg = hover_bg
        self.error_color = error_color
        self.callback = callback
        self.editable = editable
        self.dragging = False
        self.check_function = check_function
        self.display_function = display_function if display_function is not None else lambda s: s
        self.hover_enter_function = hover_enter_function
        self.hover_leave_function = hover_leave_function
        self.entry_on_function = entry_on_function
        self.entry_off_function = entry_off_function
        self.last_callback = time.time() # to avoid double callbacks

        Frame.__init__(self, master, bg=bg)
        
        self.label = Label(self, text=self.display_function(self.text),
                           bg=bg, fg=fg, font=(font_name, font_size))
        self.label.pack(fill='both')
        self.Entry = CheckEntry(self, default=self.text, allowed_chars=allowed_chars,
                                max_len=max_len, check_function=self._entry_check_callback,
                                font_name=font_name, font_size=font_size,
                                justify=justify, bg=bg, fg=fg,
                                error_color=error_color, width=width)
        self.label.bind('<Double-Button-1>', self.to_entry)
        self.label.bind("<ButtonRelease-1>", self.button_release)
        self.Entry.bind("<Return>", self.to_label)
        self.Entry.bind("<Tab>", self.to_label)
        if hover_bind:
            self.label.bind("<Enter>", self.hover_enter)
            self.label.bind("<Leave>", self.hover_leave)
        if focus_out_bind:
            self.Entry.bind("<FocusOut>", self.to_label)

    def set_active(self):
        self.editable = True

    def set_inactive(self):
        self.editable = False
        self.to_label(callback=False)

    def set_text(self, text, callback=False):
        self.Entry.pack_forget()
        self.label.focus_set() # to take focus away from Entry so that keyboard strokes are no longer read by Entry
        self.label.pack(fill='both')
        self.text = text
        self.label.config(text=self.display_function(self.text))
        self.Entry.activate(text=text, select=False, focus=False) # only to set text in entry box
        if callback and self.callback:
            self.callback(text)

    def set_bg(self, bg:str, hover_bg:str=None):
        '''updates background color and hover_bg, optionally'''
        self.bg = bg
        self.hover_bg = hover_bg if hover_bg is not None else self.hover_bg
        self.label.config(bg=self.bg)
        self.config(bg=self.bg)
        self.Entry.set_bg(self.bg)

    def set_allowed_chars(self, allowed_chars:str):
        '''updates characters allowed in entry box
        :param allowed_chars: str | list[str] | None
        '''
        self.Entry.set_allowed_chars(allowed_chars)

    def hover_enter(self, event=None):
        if self.editable:
            self.label.config(bg=self.hover_bg)
            self.config(bg=self.hover_bg)
            if self.hover_enter_function:
                self.hover_enter_function()

    def hover_leave(self, event=None):
        if self.editable and not self.dragging:
            self.label.config(bg=self.bg)
            self.config(bg=self.bg)
            if self.hover_leave_function:
                self.hover_leave_function()

    def button_release(self, event=None):
        if self.dragging:
            self.hover_leave()
            self.dragging = False

    def _entry_check_callback(self, text:str):
        '''called when text in entry box is modified'''
        check = self.check_function is None or self.check_function(text)
        return text == self.text or check

    def to_entry(self, event=None):
        if self.editable:
            self.label.pack_forget()
            self.Entry.activate(text=self.text, select=True)
            self.Entry.pack(fill='both')
            if self.entry_on_function:
                self.entry_on_function()

    def to_label(self, event=None, callback=True):
        '''switching from entry box to label'''
        current_time = time.time()
        if current_time - self.last_callback > 0.05: # avoid double callback
            self.last_callback = current_time
            self.Entry.pack_forget()
            self.label.focus_set() # to take focus away from Entry so that keyboard strokes are no longer read by Entry
            self.label.pack(fill='both')
            if self._entry_check_callback(self.Entry.get()): # only if text in Entry is good
                if self.Entry.get() != self.text: # only callback if text has changed
                    self.text = self.Entry.get()
                    self.label.config(text=self.display_function(self.text))
                    if callback and self.callback:
                        self.callback(self.text)
            if self.entry_off_function:
                self.entry_off_function()

    def get_text(self) -> str:
        '''should not be overriden'''
        return self.text

    def get(self):
        '''can be overriden in child classes to return other types'''
        return self.text

class NumberEditLabel(EditLabel):
    ''' Extension of EditLabel that only handles numbers and allows the to
        adjust the number by dragging on top of the label.

        Clicking on the label will switch it to an entry box for the user to
        enter an exact value. Dragging across the label will increment its value
        like a slider. To allow both of these functionalities, switching to the
        entry box is actually done upon the mouse button release, if the mouse
        button was released right after it was clicked (duration is defined by
        the 'drag_threshold' argument). Clicks longer than this will be
        considered drags and will not switch to the entry box.

        The sensitivity of dragging depends on the 'reference_width' parameter,
        which is 2 by default. 'reference_width' refers to a fraction of the
        label width and is defined as the the drag distance required to go from
        the minimum value to the maximum value. Dragging is entirely relative
        and does not depend on the position where the drag begun or ended.

        NumberEditLabel can handle floats or integers, depending on the value of
        the 'step' argument. If step evenly divisible by 1 (not necessarily of
        type int), return values from .get() will be integers, otherwise return
        values will be floats.
        
        NumberEditLabel is an excellent compact solution when there is not room
        to display a scrollbar or slider for the user to adjust numeric values.
    '''
    def __init__(self, master, callback=None, min_value=0, max_value=100,
                 step=1, default_value=None, reference_width=2.0, max_len=None,
                 drag_threshold=0.2, draggable=True, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param callback: function (int|float), called when value is changed
            :param min_value: int or float - minimum value
            :param max_value: int or float - maximum value
            :param step: int or float - number increment
            :param default_value: int or float - default value if different from min_value
            :param reference_width: float - fraction of label width
            :param max_len: int - maximum number of characters in entry box
            :param drag_threshold: float (seconds) - click duration to be considered a single click (not drag)
            :param draggable: bool - if True, label value can be changed by dragging (like a slider)
        '''
        self._step = step
        if self._step % 1 == 0: # integer
            self._decimals = 0
        else:
            self._decimals = len(str(self._step).split(".")[1])
        default_value = default_value if default_value is not None else min_value
        
        super().__init__(master, self._get_text(default_value),
                         callback=self._entry_update, max_len=max_len,
                         check_function=self._check_function, **kwargs)
        
        self._callback_function = callback
        self._min_value, self._max_value = min_value, max_value
        self._values = np.arange(self._min_value, self._max_value + self._step * 0.9, self._step)
        self._drag_threshold = drag_threshold
        self._reference_width = reference_width

        self._update_allowed_chars()

        if draggable:
            self.label.config(cursor='sb_h_double_arrow')
            self.label.bind("<Button-1>", self._mouse_click)
            self.label.bind("<ButtonRelease-1>", self._mouse_release, add='+')
            self.label.bind("<B1-Motion>", self._mouse_drag)

    def _check_function(self, text:str) -> bool:
        '''
        called when text is edited in entry box - checks if text is good
        input text is guaranteed to contain only numbers and '.' if return_type
        is float
        '''
        if text in ['', '-', '.']:
            return False
        if text[0] == '.' or text[-1] == '.' or text.count('.') > 1: # improperly formatted number
            return False
        if text.count('-') > 1 or ('-' in text and text[0] != '-'): # improperly formatted negative number
            return False
        value = float(text)
        if value < self._min_value or value > self._max_value: # out of range
            return False
        return True
        
    def _entry_update(self, text:str):
        '''
        called when number is updated by user exiting entry box
        text is guaranteed to have passed self._check_function
        reformats text if necessary and calls callback function
        '''
        # to reformat if necessary
        self.text = self._get_text(self.get())
        self.label.config(text=self.text)
        if self._callback_function is not None:
            self._callback_function(self.get())

    def _mouse_click(self, event):
        '''
        called when mouse clicks on label
        store current position for dragging
        store time so that release can decide to go to entry box or not
        '''
        self.dragging = True
        self._click_value = self.get()
        self._click_position = event.x
        self._click_time = time.time()

    def _mouse_release(self, event=None):
        '''
        called when mouse releases click
        if a short enough time has passed since click, goes to entry box
        a short time would indicate that the user simply clicked and didn't drag
        '''
        if time.time() - self._click_time < self._drag_threshold:
            self.to_entry()

    def _mouse_drag(self, event):
        '''
        called when mouse drags on label
        changes value based on drag distance
        '''
        # compute percentage of reference width that drag has traversed
        p = (event.x - self._click_position) / (self.label.winfo_width() * self._reference_width)
        # compute increment based on traversal percentage
        increment = (self._max_value - self._min_value) * p
        # add increment to click value and round to nearest allowed value
        value = self._values[(np.absolute(self._values - (self._click_value + increment))).argmin()]

        self.text = self._get_text(value)
        self.label.config(text=self.text)
        if self._callback_function is not None:
            self._callback_function(self.get())

    def _update_allowed_chars(self):
        '''called internally to update allowed characters in entry box
        when limits or decimals change (limits may include negative numbers)
        '''
        allowed_chars = '0123456789'
        if self._decimals > 0:
            allowed_chars += '.'
        if self._min_value < 0: # include negative numbers
            allowed_chars += '-'
        super().set_allowed_chars(allowed_chars)

    def set_min_value(self, min_value:int, force_limits=True):
        '''sets minimum value - must be less than current maximum value
        if force_limits is True, changes value and max value if necessary'''
        if force_limits:
            self._max_value = max(self._max_value, min_value) # adjust max if necessary
            if min_value > self.get():
                self.set(min_value, force_limits=False)
        self._min_value = min_value
        self._values = np.arange(self._min_value, self._max_value + self._step * 0.9, self._step)
        self._update_allowed_chars()

    def set_max_value(self, max_value:int, force_limits=True):
        '''sets maximum value - must be greater than current minimum value
        if force_limits is True, changes value and min value if necessary'''
        if force_limits:
            self._min_value = min(self._min_value, max_value) # adjust min if necessary
            if max_value < self.get():
                self.set(max_value, force_limits=False)
        self._max_value = max_value
        self._values = np.arange(self._min_value, self._max_value + self._step * 0.9, self._step)
        self._update_allowed_chars()

    def get_min_value(self) -> int:
        '''returns current minimum value'''
        return self._min_value
    
    def get_max_value(self) -> int:
        '''returns current maximum value'''
        return self._max_value

    def _get_text(self, value:int) -> str:
        '''converts value to string based on step and return_type'''
        if self._decimals == 0: # integer
            return str(int(value))
        text = str(round(value, self._decimals))
        if '.' in text:
            dec = len(text.split(".")[1])
        else:
            text += '.'
            dec = 0
        return text + '0' * (self._decimals - dec)

    def set(self, value:int, force_limits=True):
        '''updates value
        if force_limits is True, changes min and max value if necessary
        '''
        if force_limits:
            self._min_value = min(self._min_value, value)
            self._max_value = max(self._max_value, value)
        super().set_text(self._get_text(value))
        self._update_allowed_chars()

    def get(self) -> int:
        '''
        returns number current in NumberEditLabel as int or float depending
        on preset return_type
        text currently in label is guaranteed to be formatted correctly
        '''
        if self._decimals == 0: # integer
            return int(super().get())
        else:
            return float(super().get())

    def set_perc(self, perc:float):
        '''sets value based on percentage - float between 0 and 1'''
        self.set(self._min_value + (self._max_value - self._min_value) * perc)

    def get_perc(self) -> float:
        '''returns label value as a percentage - float between 0 and 1'''
        return (self.get() - self._min_value) / (self._max_value - self._min_value)

class TimeEditLabel(NumberEditLabel):
    ''' Extension of NumberEditLabel that displays time in the form
        (h):(m)m:ss.xxx instead of simple number.
        
        set() and get() methods deal in seconds.
        
        Input/output type can be int or float depending on step
        
        See NumberEditLabel for further implementation details.
    '''
    allowed_chars = '0123456789:'

    def _check_function(self, text:str) -> bool:
        '''
        called when text is edited in entry box - checks if text is good
        input text is guaranteed to contain only numbers and ':' and '.' if
        return_type is float
        '''
        if text == '':
            return False
        if text.count(':') > 2 or text[0] == ':' or text[-1] == ':':
            return False
        if text.count('.') > 1 or text[0] == '.' or text[-1] == '.':
            return False
        for t in text.split(':')[1:]: # minute, second components
            if len(t) < 2 or float(t) >= 60:
                return False
        for t in text.split(':')[:-1]: # hour, minute components
            if t == '' or '.' in t or len(t) > 2 or int(t) >= 60:
                return False
        value = self.__get_text_value(text)
        if value < self._min_value or value > self._max_value: # out of range
            return False
        return True
        
    def _get_text(self, sec:float) -> str:
        '''converts value to string based on self._decimals'''
        # compute hours, minutes, and seconds
        s = int(sec)
        if sec >= 3600: # at least one hour
            text = f'{s // 3600}:{(s % 3600) // 60:0>2}:{(s % 60) // 1:0>2}'
        else: # less than one hour
            text = f'{s // 60}:{(s % 60) // 1:0>2}'

        # compute partial seconds
        if self._decimals > 0:
           if sec % 1 == 0: # whole seconds - still add zeros
               text += '.' + '0' * self._decimals
           else:
               dec = str(round(sec % 1, self._decimals))[1:] # in form .xxx
               text += f'{dec:0>{self._decimals + 1}}'
        return text

    def __get_text_value(self, text:str) -> float:
        '''
            returns the value of text in form (h):(m)m:ss.xxx in seconds
            text passed is guaranteed to have passed self._check_function
        '''
        if text.count(':') == 2: # includes hours - in form hh:mm:ss.xxx
            hours, minutes, seconds = text.split(':')
        elif text.count(':') == 1: # includes minutes - in form mm:ss.xx
            hours = 0
            minutes, seconds = text.split(':')
        else: # only minutes - in form mm:ss.xxx
            hours, minutes = 0, 0
            seconds = text

        value = float(hours) * 3600 + float(minutes) * 60 + float(seconds)
        if self._decimals == 0: # integer
            return int(round(value, 0))
        else: # float
            return round(value, self._decimals)
        
    def set(self, seconds:float, force_limits=False):
        '''updates value - force_limits is only for consistency with NumberEditLabel'''
        self.set_text(self._get_text(seconds))

    def get(self) -> float:
        '''returns current value as int or float (seconds) depending on step'''
        return self.__get_text_value(EditLabel.get(self))

class RangeLabel(Frame):
    ''' Pair of NumberEditLabels that allow user to select a range
    '''
    def __init__(self, master, callback=None, bg='#ffffff', dropdown_bg=None,
                 sep_text='to', min_val=0, max_val=100, default_min=None,
                 default_max=None, step=1, min_range=0, label_fg='#888888',
                 label_font_size=10, **kwargs):
        '''
        Parameters
        ----------
            :param master: tk.Frame - parent widget
            :param callback: function (int, int) - called when ranges is changed
            :param bg: str (hex code) - background color
            :param set_text: str - text between the two NumberEditLabels
            :param min_val: int - minimum value
            :param max_val: int - maximum value
            :param default_min: int - default minimum value if different from min_val
            :param default_max: int - default maximum value if different from max_val
            :param step: int or float - slider increment
            :param min_range: int - minimum allowed distance between min and max
            :param label_fg: str (hex code) - color of sep_text
            :param label_font_size: int - font size of sep text
        '''
        super().__init__(master, bg=bg)
        self._min_range = min_range
        self._callback_function = callback
        default_min = default_min if default_min is not None else min_val
        default_max = default_max if default_max is not None else max_val
        max_len = len(str(max_val)) if step % 1 == 0 else None

        dropdown_bg = dropdown_bg if dropdown_bg is not None else bg
        self._MinLabel = NumberEditLabel(self, self._min_callback, bg=dropdown_bg,
                                          min_value=min_val,
                                          max_value=default_max - self._min_range,
                                          step=step, default_value=default_min,
                                          max_len=max_len, **kwargs)
        self._MaxLabel = NumberEditLabel(self, self._max_callback, bg=dropdown_bg,
                                          min_value=default_min + self._min_range,
                                          max_value=max_val, step=step,
                                          default_value=default_max,
                                          max_len=max_len, **kwargs)
        self._MinLabel.pack(side='left', fill='x', expand=True)
        Label(self, text=f' {sep_text} ', bg=bg, fg=label_fg,
              font=('Segoe UI', label_font_size)).pack(side='left')
        self._MaxLabel.pack(side='left', fill='x', expand=True)

    def set_min(self, new_min:int):
        '''updates current min value - does not change range limits'''
        self._MinLabel.set(new_min)
        self._MaxLabel.set_min_value(new_min + self._min_range)

    def set_max(self, new_max:int):
        '''updates current max value - does not change range limits'''
        self._MaxLabel.set(new_max)
        self._MinLabel.set_max_value(new_max - self._min_range)

    def set_min_limit(self, new_min_limit:int):
        '''updates min limit - does not change current values'''
        self._MinLabel.set_min_value(new_min_limit)

    def set_max_limit(self, new_max_limit:int):
        '''updates max limit - does not change current values'''
        self._MaxLabel.set_max_value(new_max_limit)

    def set_min_range(self, min_range:int):
        '''updates the minimum allowed distance between min and max values'''
        self._min_range = min_range
        self._MaxLabel.set_min_value(self._MinLabel.get() + self._min_range)
        self._MinLabel.set_max_value(self._MaxLabel.get() - self._min_range)

    def _min_callback(self, min_value:int):
        '''
        called when min label is changed
        updates minimum value of max label and calls callback function
        '''
        self._MaxLabel.set_min_value(min_value + self._min_range)
        if self._callback_function is not None:
            self._callback_function(*self.get())

    def _max_callback(self, max_value:int):
        '''
        called when max label is changed
        updates maximum value of min label and calls callback function
        '''
        self._MinLabel.set_max_value(max_value - self._min_range)
        if self._callback_function is not None:
            self._callback_function(*self.get())

    def get(self):
        '''
        Purpose:
            gets minimum and maximum value from EditLabels
        Pre-conditions:
            (none)
        Post-conditions:
            (none)
        Returns:
            :return int - minimum value
            :return int - maximum value
        '''
        return self._MinLabel.get(), self._MaxLabel.get()

class NumberIncrementLabel(Frame):
    ''' Widget that allows user to select a number by typing in CheckEntry
        or by clicking '+' and '-' buttons
        
        EditLabel is used to allow user to enter a number
        
        IconButton is used for '+' and '-' buttons
    '''
    def __init__(self, master, bg:str, fg:str, default_val:int, hover_bg='#cccccc',
                 min_val=None, max_val=None, min_val_function=None, max_val_function=None,
                 callback_function=None, font_name='Segoe UI', font_size=12,
                 max_len=None, width=4):
        '''
        
        Parameters
        ----------
            :param master: tk.Frame - widget in which to grid label
            :param bg: str (hex code) - background color
            :param fg: str (hex code) - foreground color
            :param default_val: int
            :param hover_bg: str (hex code) - background color when cursor hovers
            :param min_val: None or int - minimum value
            :param max_val: None or int - maximum value
            :param min_function: function () -> int or None - returns min value or None
            :param max_function: function () -> int or None - returns max value or None
            :param callback_function: function (int) - called when value is changed
            :param max_len: int or None - maximum number of characters in number
        '''
        self.value = default_val
        self.min_val, self.max_val = min_val, max_val
        self.min_function, self.max_function = min_val_function, max_val_function
        self.callback_function = callback_function
        Frame.__init__(self, master, bg=bg)
        self.grid_columnconfigure(0, weight=0)
        self.grid_columnconfigure(1, weight=1)
        self.grid_columnconfigure(2, weight=0)

        minus_button = IconButton(self, icons['minus'],
                                  self.minus, selectable=False,
                                  bar_height=0, popup_label='-1', inactive_bg=bg,
                                  inactive_hover_fg=fg, inactive_fg=fg,
                                  inactive_hover_bg=hover_bg)
        plus_button = IconButton(self, icons['plus'],
                                 self.plus, selectable=False,
                                 bar_height=0, popup_label='+1', inactive_bg=bg,
                                 inactive_hover_fg=fg, inactive_fg=fg,
                                 inactive_hover_bg=hover_bg)
        minus_button.grid(row=0, column=0)
        plus_button.grid(row=0, column=2)

        self.Label = EditLabel(self, str(self.value), bg=bg, fg=fg, hover_bg=hover_bg,
                               callback=self.update_from_label, check_function=self.edit_check,
                               allowed_chars='0123456789', max_len=max_len,
                               font_name=font_name, font_size=font_size,
                               justify='center', width=width)
        self.Label.grid(row=0, column=1)

    def edit_check(self, value:str):
        '''returns True if value is acceptable, otherwise False'''
        if value == '':
            return False
        value = int(value)
        if self.min_val != None and value < self.min_val:
            return False
        if self.max_val != None and value > self.max_val:
            return False
        min_exists = self.min_function != None and self.min_function() != None
        if min_exists and value < self.min_function():
            return False
        max_exists = self.max_function != None and self.max_function() != None
        if max_exists and value > self.max_function():
            return False
        return True

    def update_from_label(self, value:str):
        '''called when value is changed with label- value will have already passed edit_check'''
        self.value = int(value)
        if self.callback_function:
            self.callback_function(self.value)

    def set_value(self, value:int, callback=False):
        '''called externally or when plus or minus buttons are clicked'''
        self.value = value
        self.Label.set_text(str(self.value))
        if callback and self.callback_function:
            self.callback_function(self.value)

    def minus(self):
        if self.edit_check(self.value - 1):
            self.set_value(self.value - 1, callback=True)

    def plus(self):
        if self.edit_check(self.value + 1):
            self.set_value(self.value + 1, callback=True)

    def get(self):
        return self.value
